/*************************************************************************
  program name: lm_lods.c
  
  Program to do lod scores (replaces homozyg.c from old Autozyg)
  The state (X) consists of segregation indicators -- so can generalize to
	cases of more than one observed/pedigree, and multiple loci.
  
  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

  Obtained by modification of jnt_auto.c.  December 1998.  (EAT)

Here is the plan!
    OK this is the PLAN  (12/15/98)

OK, now we have a general autozyg program, we need to go back to
lod estimation, for those as want.  We use the same program (modified).

For k markers, we will have (k+2) simulation points, and (k+2) evaluation
points  (but maybe it is not worth evaluating all for all).
To start we will just evaluate relative to a left-point and a right-point
for each simulation point.

The simulation points are mid-point of each interval, and r=0.1 to each
end of chromosome (0.1 can be adjusted later), and "unlinked" which is
0.5 to either end.

Since the data are locus-by-locus, the likelihood ratio contribution is
just ratio of P_theta(S)/P_sim_th(S).

This is very easy, once we get the chromosome set-ups fixed.

1) Trait locus data set-up -- assume trait-locus genotypes for observed
individuals -- done  12/20/98
2) Likelihood-ratio contributions routine,  -- done 12/18/98
3) Simulation and evaluation chromosomes for each location. -- done 12/23/98

4) Need a test run!! -- JV4C maybe.   working! 12/24/98

New routines:  (12/19/98)
       full_seg_prob() and full_pen_seg_lod() and lik_rat()
          (all in full_logprob.c)
           full_seg_prob modified from old routine -- returns log(prob P(S)) 
                of all meioses on given component pedigree
           full_pen_seg_lod computes full penetrance P(Y|S) for data y, and
                  given meiosis indicators, on given component pedigree.
               calls full_lod and returns full log-probability of (Y,S).
        [lod_next routine is not currently used ]

Now to integrate the L-sampler, using the new Lsampler and Msampler
      libraries and new routines;        EAT March 2000

Converted to MORGAN_V2.3                 MAJ June 2000

Modifications to specify simulation points, within each interval and
outside of map.    EAT, then converted to MORGAN 2.3 by MAJ, Nov 2000.

Modification of full_seg_prob, and recombinant counts     EAT  Nov 2000
    scoring of counts and tot_log_probs by component and locus/interval.

Modifications to lik_rat, do_lsmpl, etc.                  EAT  Dec 2000

Modifications to specify intervals updated by the l-sampler to allow more
efficient updating of segregation probabilities in the l-ratio.
                                                          EAT Jan 2001

**************************************************************************/

#define PNAM  "lm_lods"

#include "autozyg.h"
#include "set_auto.c"
#include "quant.h"

#undef CHECK

void  chkprnt1 (int);
void  chkprnt2a (int);
void  chkprnt2b (int);
void  chkprnt5 (void);
void  chkprnt6 (void);
void  chkprnt7 (int);
void  chkprnt8 (int);
void  chkprnt9 (int);
void  chkprnt19 (int, int);
void  chkprnt10 (int);
void  chkprnt11 (int, int, int, int);
void  chkprnt12 (int);
void  chkprnt13 (int);
void  chkprnt14 (int, int);
void  chkprnt15 (int, int);
void  chkprnt16 (int);
void  chkprnt17 (void);

int ***y;				/* full genotypic array */
double l_ratio[NC][3];			/* now this is the LOG-l-ratio, EAT 12/2000 */

int   main (int argc, char *argv[])
{
    int   i, j, k, l;
    int **dodo = NULL;
    double **quant = 0;
    int   t, c, h, err;
    int   ngenes = 0, nfou, nloc;
    int  *nns, nsw, sx;
    int  *perm_sw, *perm_loc = 0;	/* permutation arrays of switches and loci */
    int   lsflag[3][NLOC];		/* flags for lsampler l_ratio adjustments */

    int   scounts[NC], recounts[NC][NLOC][10];
    double prob[NC][NLOC];
    /* likelihoods etc.        */
    double cumulik[NC][3];
    double temp_log_prob_seg[NC][NLOC];
    double temp2_log_prob_seg[NC][NLOC];
    double tot_log_prob_seg[NC][NLOC], tot_log_prob_pen[NC][NLOC];
    double tot_log_left_seg[NC][NLOC], tot_log_rght_seg[NC][NLOC];
    int   tot_count;
    int   jj;
    double rec0[3];
    int loc0[3];

    /* Matrix for likelihood ratios */
    double *** likeratios = NULL;
    double * likes = NULL;
    /* Lods scores */
    double ** lods  = NULL;
    double leftratio  = 0.0;
    double rightratio = 0.0;
    double like_eigenvalue = 0.0;

    double seg_pen2 (), full_pen_seg_lod (), lik_rat (), diff_seg_prob ();

    void  full_seg_prob ();

    void  fwd_compute (), back_smpl (), set_lsmpl (), do_msmpl ();
    void  score_genes (), set_positions ();
    void  do_lsmpl ();

    FILE *pedF, *oSeedF;

    atexit (print_tail);

    print_head (*argv);

    if (!proc_pars (argc, argv, PNAM))	/* process input parameter      */
	exit (EXIT_FAILURE);		/* statements                   */

    if (!proc_auto ())			/* check statements for         */
	exit (EXIT_FAILURE);		/* completeness and consistency */
                                        /* and prepare controls for     */
                                        /* lm_lods                      */

    if (!proc_seeds ())			/* set up seed                  */
	exit (EXIT_FAILURE);		/* parameters                   */

    if (!proc_fils ())			/* set up pedigree file         */
	exit (EXIT_FAILURE);		/* parameters                   */

    setSeeds (SAMPL);			/* initializes seeds for rans() */

    if (!chk_pedfil ())			/* checks pedigree file for     */
	give_up (ERR_qit);		/* errors.                      */

    if (!mak_opedfil ())		/* writes new pedigree file     */
	give_up (ERR_qit);		/* if changes                   */

    if (!(pedF = open_ped ()))		/*opens pedigree file           */
	give_up (ERR_qit);

    alloc_ped ();
    input_ped (pedF, 0, dodo, quant);	/*input pedigree and trait data */

    nloc = NLoci - 1;

    nfou = count_founders ();
    printf (" Number of founders:  %d\n\n", nfou);
    printf (" Number of components:  %d\n\n", n_compnts);

    MaxNsw = 2 * (pedsize - nfou);
    nns = malloc ((n_compnts + 1) * sizeof (int));
    alloc_chk (WHERE, nns);

/* now allocate space for genotypes y for our values of pedsize, nloc   */

    alloc_y (&y, nloc, pedsize);

    qdex = malloc ((pedsize + 1) * sizeof (int));
    alloc_chk (WHERE, qdex);
    for (i = 0; i <= pedsize; ++i)
	qdex[i] = 0;
    trtphen = malloc ((pedsize + 1) * sizeof (int));
    alloc_chk (WHERE, qdex);
    for (i = 0; i <= pedsize; ++i)
	trtphen[i] = 0;

/* now set marker data for observed individuals  */

    set_markdat (y, nloc, n_compnts);

/* now set trait data                            */

    set_traitdat (y, qdex, trtphen, pedsize, n_compnts);

    chkprnt1 (nloc);

    allocate (pedsize, nloc);

/*  set up initial segs config ** removed **   */

/* now set and index switches  */

    nsw = list_switches (nloc, nns);
    printf (" Number of meiosis switches to sample is %d\n", nsw);
    fflush (stdout);

/* set up permutation arrays for loci and switches */
    perm_sw = calloc (nsw, sizeof (int));
    alloc_chk (WHERE, perm_sw);
    for (i = 0; i < nsw; ++i)
	perm_sw[i] = i;
    perm_loc = calloc (nloc + 1, sizeof (int));
    alloc_chk (WHERE, perm_loc);
    for (i = 0; i <= nloc; ++i)
	perm_loc[i] = i;

/* for simplicity we specify trait locus unlinked to start
   NO--THIS CAUSES BUG!!!  EAT April 2000
   so can same set-up as lm_auto program -- but need Marek */

/* L-sampler set-up */
    set_lsmpl (0, nloc, NAll, Qq, y);	/* EAT  March 2000              */

    for (j = 0; j <= nloc; ++j)
    {
	ngenes = set_genes (j);

	chkprnt2a (j);

	for (c = 0; c < n_compnts; ++c)
	    prob[c][j] = seg_pen2 (c, y, Qq[j], j, ngenes, &err);

	chkprnt2b (j);
    }

    /* Allocate memory for likeratios */
    likes = (double *) malloc ( NTPos * sizeof(double) );
    alloc_chk(WHERE, likes);
    /* Allocate memory for lods scores */
    lods = (double **) malloc ( n_compnts * sizeof(double *) );
    alloc_chk(WHERE, lods);
    for (i = 0; i < n_compnts; ++i) {
	lods[i] = (double *) malloc ( (NTPos - 1) * sizeof(double) );
        alloc_chk(WHERE, lods[i]);
    }	

    likeratios = (double ***) malloc (n_compnts * sizeof(double **));
    alloc_chk(WHERE, likeratios);
    for (i = 0; i < n_compnts; ++i) 
    {
	likeratios[i] = (double **) malloc(NTPos * sizeof(double*));
	alloc_chk(WHERE, likeratios[i]);	
	for (j = 0; j < NTPos; ++j) 
	{
	    likeratios[i][j] = (double *) malloc(NTPos * sizeof(double));
	    alloc_chk(WHERE, likeratios[i][j]);
	}
    }
    
    					/* now start BIG loop */
    for (k = 0; k < NTPos; ++k)
    {
	printf ("\n\n CHROMOSOME SET %d\n\n", k);
	fflush (stdout);
					/* two evaluation sets for each*/
					/* simulation chromo (left/right)*/
	printf (" Recombinations:\n");

	for (sx = FEM; sx <= MAL; ++sx)
	{
            loc0[0] = IntVal[k];
            rec0[0] = RecVal[sx][k];
 
            if (k == 0)
 	    {
 		loc0[1] = nloc;
 		rec0[1] = RecVal[sx][NTPos - 1];
 	    }
 	    else
 	    {
 		loc0[1] = IntVal[k - 1];
 		rec0[1] = RecVal[sx][k - 1];
 	    }
 
 	    loc0[2] = IntVal[k + 1];
 	    rec0[2] = RecVal[sx][k + 1];

 	    set_chromosomes (nloc, Marek[sx], 2, loc0, rec0, 2 - sx);

	    for (i = 0; i <= 2; ++i)
	    {
		if (i == 0)
		    printf ("    sim,   ");
		else if (i == 1)
		    printf ("    left,  ");
		else
		    printf ("    right, ");
		if (sx == MAL)
		    printf ("M:");
		if (sx == FEM)
		    printf ("F:");
		print_chromosome (nloc, i, 2 - sx);
	    }
	}
	printf ("\n");

	chkprnt5 ();

						/* L-sampler set-up */

	set_positions (nloc);			/* EAT  March 2000              */

	chkprnt6 ();

/* now should do an L-sample step on the trait locus (0) to reset  */

	do_lsmpl (nloc, 0, prob, y, ngenes);

/*  A whole lot of uneceesary resetting junk took out of here: EAT Dec 2000 */

	chkprnt8 (k);

	for (c = 0; c < n_compnts; ++c)
	{
	    for (h = 0; h <= 2; ++h)
	    {
		l_ratio[c][h] = 0.0;
		cumulik[c][h] = 0.0;
	    }

	    for (h = 0; h <= nloc; ++h)
	    {
		tot_log_prob_seg[c][h] = 0.0;
		tot_log_left_seg[c][h] = 0.0;
		tot_log_rght_seg[c][h] = 0.0;
		tot_log_prob_pen[c][h] = 0.0;
	    }
	    scounts[c] = 0;
	    for (j = 0; j <= nloc; ++j)
	    {
		for (jj = 0; jj < 10; ++jj)
		    recounts[c][j][jj] = 0;
	    }
	}
	tot_count = 0;

	chkprnt9 (k);
	                              		/*  set initial l_ratio      */
	for (c = 0; c < n_compnts; ++c)
	{
	    l_ratio[c][1] = 0.0;
	    l_ratio[c][2] = l_ratio[c][1];

	    for (h = 1; h <= 2; ++h)
	    {
		for (j = 0; j < nloc; ++j)
		    lsflag[h][j] = 1;
		l_ratio[c][h] +=
		    diff_seg_prob (nloc, c, h, lsflag[h], temp_log_prob_seg);
	    }
	}

	chkprnt19 (n_compnts, nloc);

/* now do MCMC resampling by meiosis or by locus */

	for (t = 0; t < NIt; ++t)
	{

	    chkprnt10 (nloc);

	    if (rans () < LSamplProb)
	    {
                                                /****    Do L-sample step    ****/
		if (!((t + 1) % WIt))
		{
		    printf (" Starting L-sampler iteration %d\n", t + 1);
		    fflush (stdout);
		}
		if (Scan)		/* permute loci for L-sampler scan */
		{
		    permute (perm_loc, nloc + 1);
		    for (j = 0; j <= nloc; ++j)
		    {
			jj = perm_loc[j];

			do_lsmpl (nloc, jj, prob, y, ngenes);
		    }
		    for (h = 1; h <= 2; ++h)
			for (j = 0; j <= nloc; ++j)
			    lsflag[h][j] = 1;
		    for (c = 0; c < n_compnts; ++c)
		    {
			l_ratio[c][1] = 0.0;
			l_ratio[c][2] = l_ratio[c][1];

			for (h = 1; h <= 2; ++h)
			{
			    for (j = 0; j < nloc; ++j)
				lsflag[h][j] = 1;
			    l_ratio[c][h] +=
				diff_seg_prob (nloc, c, h, lsflag[h], temp_log_prob_seg);
			    cumulik[c][h] += exp (l_ratio[c][h]);
			}
			++scounts[c];
		    }
		}
		else
		{			/* random locus, by name */
		    jj = intran (0, nloc);
		    /* adjust adjacent intervals out of l-ratio */
		    /* note: relevant intervals may depend on h */
		    for (h = 1; h <= 2; ++h)
			for (j = 0; j <= nloc; ++j)
			    lsflag[h][j] = 0;

		    for (h = 1; h <= 2; ++h)
		    {
			j = revchro[0][jj];	/* position under sim (0) */
if ( ((j==0)&&(lpflag[1][h][j] != 0) || (lpflag[2][h][j] != 0))
   || ((j==nloc)&&(lpflag[1][h][j - 1] != 0) || (lpflag[2][h][j - 1] != 0 ))
   || ((lpflag[1][h][j] != 0) || (lpflag[2][h][j] != 0) ||
                  (lpflag[1][h][j - 1] != 0) || (lpflag[2][h][j - 1] != 0 )))

			{
			    if (j > 0)
				lsflag[h][j - 1] = 1;
			    if (j < nloc)
				lsflag[h][j] = 1;
			    if (chromo[0][j] != chromo[h][j])
			    {
				j = revchro[h][jj];
				if (j > 0)
				    lsflag[h][j - 1] = 1;
				if (j < nloc)
				    lsflag[h][j] = 1;
			    }
			}

			for (c = 0; c < n_compnts; ++c)
			    l_ratio[c][h] -= diff_seg_prob (nloc, c, h, 
                                                  lsflag[h], temp_log_prob_seg);
		    }

		    do_lsmpl (nloc, jj, prob, y, ngenes);
		    for (h = 1; h <= 2; ++h)
			for (c = 0; c < n_compnts; ++c)
			    l_ratio[c][h] += diff_seg_prob (nloc, c, h, 
                                                  lsflag[h], temp_log_prob_seg);

		    for (c = 0; c < n_compnts; ++c)
		    {
			for (h = 1; h <= 2; ++h)
			    cumulik[c][h] += exp (l_ratio[c][h]);
			++scounts[c];
		    }

		}
	    }
	    else
	    {
                                            	/****    Do M-sample step    ****/
		if (!((t + 1) % WIt))
		{
		    printf (" Starting M-sampler iteration %d\n", t + 1);
		    fflush (stdout);
		}
		if (Scan)		/* permute meioses for M-sampler scan */
		{
		    permute (perm_sw, nsw);
		    for (j = 0; j < nsw; ++j)
		    {
			l = perm_sw[j] + 1;

			i = (meiosis[l].ind).index;
			sx = meiosis[l].sex;
			c = nghd[i].component - 1;

			do_msmpl (nloc, c, i, sx, prob, y, ngenes);
		    }

		    for (h = 1; h <= 2; ++h)
			for (j = 0; j <= nloc; ++j)
			    lsflag[h][j] = 1;
		    for (c = 0; c < n_compnts; ++c)
		    {
			l_ratio[c][1] = 0.0;
			l_ratio[c][2] = l_ratio[c][1];

			for (h = 1; h <= 2; ++h)
			{
			    for (j = 0; j < nloc; ++j)
				lsflag[h][j] = 1;
			    l_ratio[c][h] += diff_seg_prob (nloc, c, h, 
                                                  lsflag[h], temp_log_prob_seg);
			    cumulik[c][h] += exp (l_ratio[c][h]);
			}
			++scounts[c];
		    }
		}
		else
					    	/*choose a random meiosis */
		{
		    l = intran (0, nsw - 1) + 1;
		    i = (meiosis[l].ind).index;
		    sx = meiosis[l].sex;
		    c = nghd[i].component - 1;

		    chkprnt11 (l, i, sx, c);

		    chkprnt12 (c);

		    /* adjust meiosis to be updated out of l_ratio */

		    chkprnt13 (c);

		    for (h = 1; h <= 2; ++h)
			l_ratio[c][h] -= lik_rat (nloc, i, sx, h);

		    chkprnt14 (i, sx);

		    				/*  do-msmpl */
		    do_msmpl (nloc, c, i, sx, prob, y, ngenes);

		    /* put updated meiosis back in to l_ratio and
		     * update  the cumulative ratio                */

		    chkprnt15 (i, sx);

		    for (h = 1; h <= 2; ++h)
		    {
			l_ratio[c][h] += lik_rat (nloc, i, sx, h);
			cumulik[c][h] += exp (l_ratio[c][h]);
		    }
		    ++scounts[c];

		    chkprnt16 (t);
		}
	    }
	    if (!((t + 1) % ScIt))
	    {				/* score tot log-probs and recombs */
		for (c = 0; c < n_compnts; ++c)
		{
		    for (j = 0; j <= nloc; ++j)
			tot_log_prob_pen[c][j] += log (prob[c][j]);

		    full_seg_prob (nloc, c, recounts, temp_log_prob_seg);
		    for (j = 0; j < nloc; ++j)
			tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];

		    for (h = 1; h <= 2; ++h)
			for (j = 0; j < nloc; ++j)
			    lsflag[h][j] = 1;

		    (void) diff_seg_prob (nloc, c, 1, lsflag[1], 
                                                            temp2_log_prob_seg);
		    for (j = 0; j < nloc; ++j)
			tot_log_left_seg[c][j] += temp2_log_prob_seg[c][j] +
			    temp_log_prob_seg[c][j];

		    (void) diff_seg_prob (nloc, c, 2, lsflag[2], 
                                                            temp2_log_prob_seg);
		    for (j = 0; j < nloc; ++j)
			tot_log_rght_seg[c][j] += temp2_log_prob_seg[c][j] +
			    temp_log_prob_seg[c][j];
		}
		++tot_count;
	    }

	}

/* now we are done with this simulation/estimation for this trait locus
   position  */
/* output results for this simulation.estimation set */

	for (c = 0; c < n_compnts; ++c)
	{
	    printf ("\n Component %d:\n\n", c + 1);
	    if (scounts[c] == 0)
                printf ("    No likelihood ratios because sampling count = 0\n\n"); 
            else
            {
	        leftratio  = cumulik[c][1] / scounts[c];
	        rightratio = cumulik[c][2] / scounts[c];
	        printf ("    Cumulative likelihood ratios (sampling count %d)\n",
                                       scounts[c]);
	        printf ("       estimated ratios:   left:sim = %6.4f,"
		        "  right:sim = %6.4f\n\n", leftratio, rightratio);
            } 
	    /* Fill in the matrices of likehood ratios */
	    /* Initialization first */
	    for (j = 0; j < NTPos; ++j)
		likeratios[c][k][j] = 0.0;
	    if ( k == 0 ) 
	    {
		likeratios[c][0][NTPos-1] = leftratio;
		likeratios[c][0][1]       = rightratio;
	    }
	    else if ( k == NTPos - 1 )
	    {
		likeratios[c][k][k-1]     = leftratio;
		likeratios[c][k][0]       = rightratio;
	    }
	    else
	    {
		likeratios[c][k][k-1]     = leftratio;
		likeratios[c][k][k+1]     = rightratio;
	    }

	    printf ("    Average total log-probability of segregations\n");
	    printf ("       first locus (constant): %7.3f\n\n",
	                   	    -nns[c + 1] * log (2.0));
            printf ("       marker interval    left          sim       right\n");
	    for (j = 0; j < nloc; ++j)
                printf("            %4d      %8.3f     %8.3f    %8.3f\n",
                        j + 1, tot_log_left_seg[c][j] / tot_count,
	         	       tot_log_prob_seg[c][j] / tot_count,
			       tot_log_rght_seg[c][j] / tot_count);
	    printf ("\n    Average total log-probability of penetrances\n");

            printf ("       trait locus: %7.3f\n", tot_log_prob_pen[c][0] / tot_count);
            printf ("       marker loci: ");
	    for (j = 1; j <= nloc; ++j)
		printf ("%7.3f", tot_log_prob_pen[c][j] / tot_count);
	    printf ("\n\n");

	    fflush (stdout);
	    				/* print out recombination counts  */
	    printf ("    Recombination counts:\n");
	    printf ("       marker interval   M:  total  recomb non-rec    "
		    "F:  total  recomb non-rec\n");
	    for (j = 0; j < nloc; ++j)
	    {
		printf ("               %d          ", j + 1);
		for (i = 0; i < 6; ++i)
		{
		    if (i == 3)
			printf ("     ");
		    printf ("%8d", recounts[c][j][i]);
		}
		printf ("\n");
	    }
	    fflush (stdout);
	}
    }
					/*****    end of BIG loop   *****/

/* now we are done with simulating for all nloc+2 simulation sets */

    fflush (stdout);

    printf("\n\n ESTIMATED LODS SCORES\n");

#ifdef DEBUG
    /* Out put likehood ratio matrices */
    printf("\n  Printing likelihood ratio matrices.\n");
    printf("  Number of positions = %d\n", NTPos);
    for (c = 0; c < n_compnts; ++c) {
	printf("\n   Component %d\n", c+1);
	for (i = 0; i < NTPos; ++i) {
	    for (j = 0; j < NTPos; ++j) {
		printf("%8.4f", likeratios[c][i][j]);
	    }
	    printf("\n");
	}
    }
#endif /* DEBUG */

    /* Call like_eigen */ 
    printf("\n    The largest eigenvalue is:\n");
    for (c = 0; c < n_compnts; ++c) {
	like_eigenvalue = lik_eigen(likeratios[c], likes, NTPos);
	printf("       component %d %8.4f\n", c+1, like_eigenvalue);
	for (i = 0; i < NTPos - 1; ++i) {
            lods[c][i] = log10(likes[0]/likes[i+1]);
        }		
    }

    printf("\n    LOD scores:\n");
    printf("       trait position");
    for (c = 0; c < n_compnts; ++c) {
	printf("  cmpnt%3d", c+1);
    }
    printf("\n");
    for (i = 0; i < NTPos - 1; ++i) {
        printf("        %7d      ", i+1);
        for (c = 0; c < n_compnts; ++c) {
            printf("  %8.4f", lods[c][i]);
        }
        printf("\n");
    } 
    printf("\n");

    /* Free the memory */
    for (i = 0; i < n_compnts; ++i) {
	for (j = 0; j < NTPos; ++j) {
	    free( (void*) likeratios[i][j] );
	    likeratios[i][j] = NULL;
	}
	free ( (void**) likeratios[i] );
	likeratios[i] = NULL;
	free ( (void*) lods[i] );
	lods[i] = NULL;
    }
    free ( (void***) likeratios );
    likeratios = NULL;
    free ( (void**) lods );
    lods = NULL;
    free ( (void*) likes );
    likes = NULL;

    /* Free up allocated memory */
    free_allocate ();
    free_gen_dat ();
    for (j = 0; j <= nloc; j++)
    {
	free (y[0][j]);
	free (y[1][j]);
    }
    free (y[0]);
    free (y[1]);
    free (y);
    free_allocate_loci ();

    locus_peel_free ();

    chkprnt17();

    /* now we are done  */

    if (SamplSeedsOut)
    {
	getSeeds (SAMPL);
	if (!(oSeedF = open_fil (OSEED)))
	    give_up (ERR_qit);
	writSeeds (oSeedF);
	fclose (oSeedF);
    }

    exit (EXIT_SUCCESS);

}

/****************************************************************/

void  chkprnt1 (int nloc)
{
#ifdef CHECK
    int   i, j;
/* Check out all the data -- trait genotypes and markers   */
    for (i = 0; i < pedsize; ++i)
    {
	printf (" ind = %d", nghd[i].self);
	for (j = 0; j <= nloc; ++j)
	    printf ("    %d %d", y[0][j][i], y[1][j][i]);
	printf ("\n");
    }
    printf ("\n");
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt2a (int j)
{
#ifdef CHECK
    printf (" Set genes at locus %d\n", j);
    fflush (stdout);
    print_obs_genes (j, y);
    fflush (stdout);
    printf (" Scored prob at locus %d; giving", j);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt2b (int j)
{
#ifdef CHECK
    int   c;
    for (c = 0; c < n_compnts; ++c)
	printf (" %7.4f", prob[c][j]);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt5 (void)
{
#ifdef CHECK
    printf (" Now to call set_positions\n\n");
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt6 (void)
{
#ifdef CHECK
    printf (" Now did call set_lsampler\n\n");
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt7 (int j)
{
#ifdef CHECK
    printf (" Reset trait locus meioses, position %d\n\n", j);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt8 (int k)
{
#ifdef CHECK
    printf (" Done set-up %d \n\n", k);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt9 (int k)
{
#ifdef CHECK
    printf (" Initial seg lhood-ratios for simulation chromosome %d\n\n", k);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt19 (int n_compnts, int nloc)
{
#ifdef CHECK
    int   c;
    for (c = 0; c < n_compnts; ++c)
    {
	printf (" Component %d", c + 1);
	printf ("    left point %7.4f,    right point %7.4f\n\n",
		l_ratio[c][1], l_ratio[c][2]);
	fflush (stdout);
    }
#endif /*CHECK */
    return;
}
void  chkprnt10 (int nloc)
{					/* print set-up if needed for chk */
#ifdef CHECK
    int   j;
    for (j = 0; j <= nloc; ++j)
    {
	printf (" Locus %d position %d set-up \n", j, chromo[0][j]);
	print_set_up (j);
	printf ("\n");
	fflush (stdout);
    }
    printf ("\n");
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt11 (int l, int i, int sx, int c)
{
#ifdef CHECK
    printf (" sw %d, ind %d %d, cmp %d\n\n",
	    l, nghd[i].self, sx, c);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt12 (int c)
{
#ifdef CHECK
    int   j, h;
    for (j = 0; j <= nloc; ++j)
	printf ("\n log_pen_prob locus %d is %7.4f", j, log (prob[c][j]));
    for (h = 1; h <= 2; ++h)
	printf ("\n chromoset %d log_segs  %7.4f, cumulative totals %7.4f",
		h, diff_seg_prob (nloc, c + 1, h),
		tot_log_prob_seg[c][h]);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt13 (int c)
{
#ifdef CHECK

    printf ("\n chk l-ratio = %7.4f %7.4f", l_ratio[c][1],
	    l_ratio[c][2]);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt14 (int i, int sx)
{
#ifdef CHECK
    printf ("\n chk adj l-ratio by %7.4f %7.4f", lik_rat (nloc, i, sx, 1),
	    lik_rat (nloc, i, sx, 2));
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt15 (int i, int sx)
{
#ifdef CHECK
    printf (" chk adj l-ratio back by %7.4f %7.4f\n",
	    lik_rat (nloc, i, sx, 1),
	    lik_rat (nloc, i, sx, 2));
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt16 (int t)
{
#ifdef CHECK
    printf (" iter %d:  lrat left %7.4f, lrat rght %7.4f\n\n",
	    t + 1, l_ratio[c][1], l_ratio[c][2]);
    fflush (stdout);
#endif /*CHECK */
    return;
}
void  chkprnt17 (void)
{
#ifdef CHECK
    printf (" \n Now we are done freeing\n\n ");
    fflush (stdout);
#endif /*CHECK */
    return;
}
