
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*   nghds.h:  neighborhood structures used by polyEM programs         */
/*             also macros and function prototypes for Nghds/ fncs     */
/*             including data interfacing between Pars/ proc_poly()    */
/*             and Nghds/ fncs                                         */
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef NGHDS_H
#define NGHDS_H

#include "incs.h"

typedef struct _Ind {
	struct _Ind *pa,*ma;
	int self;
	int index;
	int component;
	struct _Marlist *marriages;
} Ind;

/*      int ntrips;        removed from _Ind  */
/*      int *trips;        used by PBGibbs    */
/*      int nbips;         (MORGAN V1.1)      */
/*      int *bips;                            */

typedef struct _Component {
	int index;
	struct _Indlist *people;
} Component;

typedef struct _Mar {
	struct _Ind *spouse;
	struct _Indlist *kids;
} Mar;

typedef struct _Marlist {
	struct _Mar *current;
	struct _Marlist *next;
} Marlist;

typedef struct _Indlist {
	struct _Ind *current;
	struct _Indlist *next;
} Indlist;
                                        /* parameters set by Nghds/     */ 
                                        /*   functions:                 */

GLOBAL int pedsize;                     /* no. of individuals read      */
GLOBAL int namestring;                  /* largest name in pedigree file*/
GLOBAL int n_compnts;			/* # of components found	*/
GLOBAL int max_cpsize;  		/* maximum component size found */

GLOBAL int **posit;	                /* was int posit[MAX_NAMES][2]  */
GLOBAL int **data;	                /* was data[MAX_SIZE][7]        */

GLOBAL Ind *nghd;	   
GLOBAL Component *cmplist;              /* vector of components         */
GLOBAL Ind nullnode;

void alloc_ped(void);
void alloc_pos(void);
void free_ped(void);
void free_nghds(void);
void free_cmplists(void);
void position(int **, int **);
void nghds(int **);
void input_ped(FILE *, int, int **, double **);
int neighbour(int, int);
int infect(int);
int components(void);

int get_posit(int);
int count_founders(void);
int intmax(int *aa, int n1, int n2);

double kinship(Ind *, Ind *),     inbreeding(Ind *);

/* Exact IBD status calculation on two linked loci through recursive
   formula. */
double g4link(int, int, int, int, int, int, int, int,
              double /* Recombination fraction */
              );

#endif  /*NGHDS_H*/

