/*********************************************************************
   Author     : Michael Na Li <nali@stat.washington.edu>
   Created    : Thu Aug 17 13:44:39 PDT 2000
   Modified   : $Date: 2000/09/29 20:02:51 $
   Version    : $Revision: 1.2 $
   CVS        : $Id: g4link.c,v 1.2 2000/09/29 20:02:51 nali Exp $
**********************************************************************/

#include "nghds.h"

double g4link (int a,           /* a is one parent of ta, and so on */
	       int b,  
	       int c,
	       int d,
	       int ta,          /* Four individuals (could be the same) */
	       int tb, 
	       int tc, 
	       int td,
	       double r         /* Recombination fraction between loci */
	       )
/* recursive program for identity at two linked loci  */
{
  int aa, bb, cc, dd, mm, ff, taa, tbb, tcc, tdd; 
  int aa_index;                         /* The expandee's index */
  double xx = 0.0;

  double q = 1.0 - r;
  double rr = 0.5 - r * q;

  if ((a*b*c*d) == 0 ) { /* One of them is a founder */
    return 0.0;
  } 
  
  aa = max(a, b); 
  bb = min(a, b); 
  cc = max(c, d); 
  dd = min(c, d);
 
  if (aa == a ) { 
    taa = ta; 
    tbb = tb; 
  } else { 
    taa = tb; 
    tbb = ta; 
  }
  
  if (cc == c ) { 
    tcc = tc; 
    tdd = td; 
  } else { 
    tcc = td; 
    tdd = tc; 
  }
  
  if (aa < cc) { 
    aa = cc; 
    taa = tcc; 
    bb = dd; 
    tbb = tdd;
    cc = max(a, b); 
    dd = min(a, b); 
    if (cc == a ) { 
      tcc = ta; 
      tdd = tb; 
    } else { 
      tcc = tb; 
      tdd = ta; 
    } 
  }
  
  aa_index = posit[aa][0];
  mm = nghd[aa_index].ma->self; 
  ff = nghd[aa_index].pa->self;    
  /* all now set so aa is expandee */
  
  if (aa == cc) {
    if ((ff == 0) && (mm == 0)) {
      if ((aa != bb) || (aa != dd)) {
	return 0.0;
      } else {
	if (((taa == tcc) && (tbb == tdd )) || ((taa == tdd) &&(tbb == tcc))) {
	 return rr;
	} else { 
	  return 0.25; 
	} 
      } /* founder expandee a */ 
    } else if ((aa == bb) && (aa == dd)) { 
      if (((taa == tcc) && (tbb == tdd )) || ((taa == tdd) &&(tbb == tcc))) {
	xx = (rr * (1.0 + g4link(mm, ff, mm, ff, aa, aa, aa, aa, r)) 
	      + 2.0 * r * q * kinship(nghd+posit[mm][0], nghd+posit[ff][0]));
     } else {
       xx = (0.25 * (1.0 + g4link(mm, ff, mm, ff, aa, aa, aa, aa, r)) 
	     + 0.5 * kinship(nghd+posit[mm][0], nghd+posit[ff][0]));
     }
     return xx; 
   } else if ((aa != bb) && (aa != dd)) {   
     if (taa == tcc) {
       xx = 0.5 * (q * (g4link(mm, bb, mm, dd, aa, tbb, aa, tdd, r)  +
			g4link(ff, bb, ff, dd, aa, tbb, aa, tdd, r)) +
		   r * (g4link(mm, bb, ff, dd, aa, tbb, aa, tdd, r) + 
			g4link(ff, bb, mm, dd, aa, tbb, aa, tdd, r)));
     } else {
       xx = 0.25 * (g4link(mm, bb, mm, dd, aa, tbb, aa, tdd, r) +
		    g4link(ff, bb, ff, dd, aa, tbb, aa, tdd, r) +
		    g4link(mm, bb, ff, dd, aa, tbb, aa, tdd, r) +
		    g4link(ff, bb, mm, dd, aa, tbb, aa, tdd, r));
     }
     return xx; 
   } else { 
     dd = min (bb,dd);   /* now have dd as only one != aa */
     if (dd == bb) {
       tdd = tbb; 
     }
     xx = 0.25 * (kinship(nghd+posit[mm][0], nghd+posit[dd][0]) +
		  kinship(nghd+posit[ff][0], nghd+posit[dd][0]) + 
		  g4link(mm, ff, mm, dd, aa, aa, aa, tdd, r) +
		  g4link(mm, ff, ff, dd, aa, aa, aa, tdd, r));
     return xx; 
   }
  } else {    
    if (aa == bb) { 
      if ((mm == 0) && (ff == 0)) {
	return 0.5 * kinship(nghd+posit[cc][0], nghd+posit[dd][0]); 
      } else { 
	xx = (0.5 * (kinship(nghd+posit[cc][0], nghd+posit[dd][0]) + 
		     g4link(mm, ff, cc, dd, aa, aa, tcc, tdd, r)));       
	return xx; 
      }
    } else {
      if ((mm == 0) && (ff == 0)) {
	return 0.0; 
     } else { 
       xx = ( 0.5 * (g4link(mm, bb, cc, dd, aa, tbb, tcc, tdd, r) + 
		     g4link(ff, bb, cc, dd, aa, tbb, tcc, tdd, r)));
       return xx; 
     }
    }
  }
}
