#include "parseauto.h"

/*
 * proc_auto() has as its input the results of parsing the parameter
 * statements by proc_pars() for the Autozyg programs.  It checks this
 * data for consistency and completeness and, if there are no errors,
 * builds data structures used by lm_auto.
 *
 * this code is derived from proc_drop 
 * one chromosome only is selected
 * there must be at least one marker
 * one trait only (linked or unlinked for lm_auto, positioned along
 * the chromosome for lm_lods) is expected.
 * any combination of sex-not-specified maps or sex-specific maps
 * may be supplied
 *                                                          maj 5/2000 */


int proc_auto(void)
{                            
    int lnkd, m_no, entry, sex;
    int sxsp_map = FALSE;
    SIM_PTR     p_s;
    DIST_PTR    p_d; 	
    UNLNKD_PTR  p_u;
    TLIST_PTR   p_t;
    VMAP_PTR    p_v;
    PRNTMAP_PTR p_m;
    PRNTLOC_PTR p_l;
					
                                        /* checks for select statement */
    if (!test_atz_sel())	        
        if (proc_err())
            return FALSE;
    if (n_err)
        return proc_end();

    entry = FALSE;
    p_s = nxt_sim_chrom(&entry);        /* selects one chromosome only  */

                                        /* checks marker map & trait map*/
                                        /* (sex-averaged & sex-specific)*/
    if (!test_mark_map(p_s, &sxsp_map))
        if (proc_err())
            return FALSE;
    if (pmx->pcod == 'a')  
        if (!test_trait_map(p_s, &sxsp_map))	
            if (proc_err()) 	  
                return FALSE;
    if (pmx->pcod == 'l')
        if (!test_trait_vmap(p_s, &sxsp_map))	
            if (proc_err()) 	  
                return FALSE;
    if (n_err)                          /* if errors, quit              */
        return proc_end();

                                        /* adds mapped trait to trait   */ 
    lnkd = TRUE;		        /* list  (lm_auto)              */
    entry = FALSE;   
    if (p_d = nxt_mapd_trait(&entry, FEM, p_s))
        if (!test_trait_list(lnkd, p_d->id))
            if (proc_err()) 
                return FALSE;
 			                /* adds unlinked trait to trait */
    lnkd = FALSE;	         	/* list  (lm_auto)            	*/
    entry = FALSE;		                               	  
    if (p_u = nxt_unlnkd_trait(&entry))
        if (!test_trait_list(lnkd, p_u->id))
            if (proc_err())   
                return FALSE;
                                        /* adds trait with varying posi-*/
    lnkd = FALSE;		        /* tions to trait list (lm_lods)*/
    if (p_v = nxt_trait_vmap(FEM, p_s))
        if (!test_trait_list(lnkd, p_v->id))
            if (proc_err()) 
                return FALSE;
					/* for each marker,		*/
    entry = FALSE;	                /* check allele frequencies     */
    while (m_no = nxt_mark(&entry, p_s))  
        if (!test_mark_loci(p_s, m_no))  
            if (proc_err())   
                return FALSE;
					/* for the trait,		*/
    entry = FALSE;			/* checks allele frequencies	*/
    if (p_t = nxt_trait(&entry))
        if (!test_trait_loci(p_t->id))
            if (proc_err())
                return FALSE;
    
    if (n_err)			
        return proc_end();
				        /* if no errors, prints out 	*/
                   			/* parameters	                */
    for (sex = FEM; sex <= AVG; ++sex)
    {                                   /* first assigns loci to sex-   */
                                        /* specific print maps          */
                                        /* defaults to sex-averaged,    */
                                        /* then to F maps, for maps not */
                                        /* present                      */

                                 	/* puts markers into map	*/
        p_m = add_prntmap(p_s, sex);           

        if (p_s->no_mark == 1)	
            add_only_mark_locus(p_m);
        else
        {
            entry = FALSE;                 
            while (p_d = nxt_mapd_mark(&entry, sex, p_s))  
                add_mark_locus(p_m, p_d);              
        }
                   		        /* adds traits to map  (lm_auto)*/
        entry = FALSE;	
        if (p_d = nxt_mapd_trait(&entry, sex, p_s))   
            add_trait_locus(p_m, p_d);
				        /* finds distances between 	*/
        entry = FALSE;          	/* adjacent loci		*/ 
        while (p_l = nxt_prntmap_locus(&entry, p_m)) 
            fix_locus_dist(p_m, p_l);             
  
        if ((sxsp_map && (sex == FEM || sex == MAL))
                 || (!sxsp_map && sex == AVG))
        {
            prnt_map(p_m, sex);         /* prints the map               */

                                        /* checks trait distances 	*/
            entry = FALSE;	  	/* (lm_auto)                    */
            while (p_l = nxt_prntmap_locus(&entry, p_m)) 
                if(!test_trait_dist(p_m, p_l))
                    if (proc_err())
                        return FALSE;
                                        /* prints the trait intervals 	*/
                                        /* (lm_lods)                    */
            if (p_v = nxt_trait_vmap(sex, p_s))        
                prnt_vmap(p_v, sex); 
        }
    } 

    prnt_mark_freqs(p_s);         	/* prints marker allele freqs   */

    prnt_trait_freqs(p_m);  	        /* prints trait allele freqs 	*/ 

    prnt_unlnkd_freqs();

    prnt_lods_freqs(p_s);

    fprintf(stdout,"\n");
   
    if (n_err)			
        return proc_end();

    NLoci = 0;		        	/* counts loci           	*/
    NMarks = p_m->no_mark;
    NTraits = p_m->no_traits;
    NUnlnkd = 0;

    if (p_u || p_v)    
    {
        ++NTraits;
        ++NUnlnkd;
    }
    NLoci = NMarks + NTraits;


    if (pmx->pcod == 'a')               /* stores map info for each     */
    {                                   /* locus in chromo[][] (lm_auto)*/
        mak_chromo(p_m);          
                                        /* stores recombination         */
                                        /* fractions in xecomo[][][]    */
        for (sex = FEM; sex <= MAL; ++sex)  
        {
            entry = FALSE;
            if (p_m = nxt_prntmap(&entry, sex))
                mak_xecomo(p_m, 2-sex); /* uses [1] male, [2] female    */
        }                               /* as subscripts                */
    }

    if (pmx->pcod == 'l')               /* stores marker & trait recomb */
    {                                   /* fracts in Marek[][] &        */
                                        /* in RecVal[][] & loci in      */
                                        /* IntVal[]  (lm_lods)          */

                                        /* sets NTPos, the no. of trait */
                                        /* positions.  (Arrays are 1    */
                                        /* greater)                     */
        NTPos = 2 * p_v->no_xtrn + (NMarks - 1) * p_v->no_prop + 1;

        alloc_lods_arrs();

        mak_intval(p_v);

        for (sex = FEM; sex <= MAL; ++sex)    
        {
            entry = FALSE;
            if ((p_m = nxt_prntmap(&entry, sex)) &&
                   (p_v = nxt_trait_vmap(sex, p_s)))
                mak_marek(p_m, sex);
                mak_recval(p_m, p_v, sex);
        }
    }
     			                /* for each locus, stores       */
    entry = FALSE;                      /* allele frequencies in Qq[][] */
    p_m = nxt_prntmap(&entry, FEM);
    entry = FALSE;           
    while (p_l = nxt_prntmap_locus(&entry, p_m)) 
        mak_allefrq(p_m, p_l, p_l->id);                
    if (p_u)            
        mak_allefrq(NULL, NULL, p_u->id);
    if (pmx->pcod == 'l')              
        mak_allefrq(NULL, NULL, p_v->id);

    if (!test_markdat(p_s))             /* checks marker data           */
        if (proc_err())
            return FALSE;
    if (!n_err)
    {
        mak_markdat(p_s);               /* stores marker data           */
        prnt_markdat(p_s);
    }

    if (pmx->pcod == 'a')       
    {                                   /* checks proband gametes       */
        if (!test_probgam())          
            if (proc_err())
                return FALSE;
        if (!n_err)
        {                               /* stores proband gametes       */
            mak_probgam();             
            prnt_probgam();    
        }
    }

    if (!test_atz_pars())               /* checks computational parms   */
        if (proc_err())
            return FALSE;

    if (!test_atz_inpt())               /* checks input parmeters      */
        if (proc_err())
            return FALSE;

    if (n_err)			
        return proc_end();

    prnt_atz_pars();                    /* prints program parameters    */

    mak_atz_inpt();      		/* set up stuff for reading	*/

    return TRUE;
}

/* ----- test fncs:  test results of processing input statements ------ */

/* -------------------- test functions for autozyg -------------------- */


VMAP_PTR nxt_trait_vmap(int sx, SIM_PTR p_sim)
{				  	/* returns the pointer to the   */
					/* trait map for the selected   */
                                        /* chromosome and gender        */
    int i, j;
    VMAP_PTR p, p_vmap = NULL;

    for (i = 1; !p_vmap && i <= 4; ++i)
    {
        if (i == 1)			/* checks for gender-specific	*/
        j = sx;				/* map present in this order    */ 
        else if (i == 2)
        j = AVG;
        else if (i == 3)
        j = FEM;
        else if (i == 4)
        j = MAL;
    
        if (p = first_trait_vmap_node[j])
        {
            while (p->p_next && (p->chrom_no < p_sim->chrom_no))
                p = p->p_next;
            if (p->chrom_no == p_sim->chrom_no)
                p_vmap = p;
        }
    }
    return p_vmap;
}

int test_atz_sel(void)		
{		                        /*checks select statements      */ 
    int n;       
    UNLNKD_PTR p;

    if (!first_sim_node)                /*select ... markers .. required*/
        err_typ = ERA_SEL; 

    else if (first_sim_node->p_next)    /*checks for only one chromosome*/
        err_typ = ERA_OOS; 

    else 
    {                                   /*checks for one mapped or      */
        n = first_sim_node->no_traits;  /*unlinked trait                */

        if (p = first_unlnkd_node) 
        {
            n++; 
            while (p = p->p_next)
                n++; 
        }
        if (n == 0) 
        {    
            if (pmx->pcod == 'a')  
                err_typ = ERA_OOT; 
            else
                err_typ = ERA_OLT;
        }
        if (n > 1)
        {
            sprintf(errnumb, "%d", n);
            err_txt = errnumb;
            err_typ = ERA_MOT;
        }
    }

    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

int test_markdat(SIM_PTR p_sel)
{ 
    int i, j, ix;
    MARKDAT_PTR p_mdt;
    LOCUS_PTR p_locus;
                                        /* finds marker data for this   */
    if (!(p_mdt = first_markdat_node))  /* chromosome                   */
    {
        NObs = 0;
        err_typ = ERA_NMD; 
        for_chrom(p_sel->chrom_no);
    }
    else
    {
        while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
            p_mdt = p_mdt->p_next;
        if (p_mdt->chrom_no != p_sel->chrom_no)
        {
            NObs = 0;
            err_typ = ERA_NMD; 
            for_chrom(p_sel->chrom_no);
        }
        else				/* checks no. of data		*/
        {
            NObs = p_mdt->no_ints/(2 * NMarks + 1);
            if (p_mdt->no_ints % (2 * NMarks + 1))
            {
                err_typ = ERA_BND;
                for_chrom(p_mdt->chrom_no);
            }
        }
    }
    if (!err_typ)                 	/* tests for duplicate names	*/
        for (i = 0; i < NObs - 1; ++i)
            for (j = i + 1; j < NObs;  ++j)
                if (*(p_mdt->p_list + i * (2 * NMarks + 1)) ==
                    *(p_mdt->p_list + j * (2 * NMarks + 1)))
                {
                    sprintf(errnumb, "%d", 
                               *(p_mdt->p_list + i * (2 * NMarks + 1)));
                    err_txt = errnumb;
                    err_typ = ERA_DIN;
                }

    if (!err_typ)
    {					/* tests marker alleles within	*/
        ix = 0;				/* range ...            	*/
        for (i = 0; i < NObs; ++i)      /* for each observed individual */
        {
            p_locus = first_mark_locus_node;
            ++ix;
            for (j = 0; j < NMarks; ++j)        /* and each marker      */
            {
                while ((p_locus->p_next 
                                && (p_locus->chrom_no < p_sel->chrom_no)
                       || (p_locus->chrom_no == p_sel->chrom_no
                                && p_locus->id < j + 1)))
                    p_locus = p_locus->p_next;

                if (*(p_mdt->p_list + ix) > p_locus->no_alleles 
                       || *(p_mdt->p_list + ix + 1) > p_locus->no_alleles)
                {
                    sprintf(errnumb, "%d", j + 1);
                    err_txt = errnumb; 
                    err_typ = ERA_WMA;
                }
                ix += 2;
            }
        }
    }
    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

int test_probgam(void)
{
    PROBGAM_PTR p_pg;

    if (!(p_pg = first_prob_gam_node))
        err_typ = ERA_MPG;
					
    if (err_typ)
        return FALSE;
    else
        return TRUE;
}
        
int test_atz_pars(void)       
{
    fprintf(stdout," \n");


    if (!NIt)
    {
       err_typ = ERA_NIT;
       return FALSE;
    }
    if (!WIt)
        WIt = NIt;
        
    if (!FWIt)
        FWIt = NIt;
        
    if (TraitPheno == UNDEF)
        TraitPheno = N;

    if (Scan == UNDEF)
        Scan = N;

    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

int test_atz_inpt(void)
{
    if (NInts == 0 && NReals == 0)
    {
        err_typ = ERF_IPR;
        return FALSE;
    }
                                       /* allows for gender added when	*/
                                       /* pedigree checking		*/ 
    else if (first_tlist_node->id + 4 > NInts + (SexIn == N ? 1 : 0))
    {
        err_typ = ERA_NEI;
        sprintf(errnumb, "%d", first_tlist_node->id);
        err_txt = errnumb;
        return FALSE;
    }
    return TRUE;
}

int test_trait_vmap(SIM_PTR p_sim, int * ss_map)
{                                       /*tests for trait map present   */
                                        /*(may be gender-specific)      */
                                        /*for this chromosome and       */
                                        /*correctly specified           */
    int sx, j;                
    VMAP_PTR p_vmap, p_m[3] = {NULL, NULL, NULL};                     

				
    if (p_sim->no_traits != 1)		/*checks for one trait only	*/
    {
            err_typ = ERA_NAT;
            return FALSE;
    }
                                        /*checks for sex-specific maps  */
    for (sx = FEM; sx <= AVG; ++sx)     /*present                       */
        if (p_vmap = first_trait_vmap_node[sx])
        {
            while (p_vmap->p_next  &&  p_vmap->chrom_no < p_sim->chrom_no)
                p_vmap = p_vmap->p_next;
            if (p_vmap->chrom_no == p_sim->chrom_no)
            { 				
                if (sx != AVG)	
                    *ss_map = TRUE;        
                p_m[sx] = p_vmap;
            }
        }
                                        /*sets defaults for sex-specific*/
                                        /*trait maps not present        */
                                        /*checks for at least one map   */
    for (sx = FEM; sx <= AVG; ++sx)     /*present                       */
        if (!p_m[sx])
            if (!(p_m[sx] = p_m[AVG]))
                if (!(p_m[sx] = p_m[FEM]))
                    if(!(p_m[sx] = p_m[MAL]))
                        if (p_sim->no_mark > 1)
                	    err_typ = ERA_MBM;
                        else 
                	    err_typ = ERA_MXM;
    if (err_typ)
    {
        for_chrom (p_sim->chrom_no);

        return FALSE;
    }
					/*checks for sex-specific       */
                                        /*external intervals and        */
                                        /*internal interval proportions */
                                        /*present                       */
    for (sx = FEM; sx <= AVG; ++sx)
    {
        if (!p_m[sx]->no_xtrn)
        {
            j = sx;
            if (p_m[AVG]->no_xtrn)
                j = AVG;
            else if (p_m[FEM]->no_xtrn)
                j = FEM;
            else if (p_m[MAL]->no_xtrn)
                j = MAL;
            else
                err_typ = ERA_MXM;

            p_m[sx]->no_xtrn = p_m[j]->no_xtrn;
            p_m[sx]->p_dist = p_m[j]->p_dist;
            p_m[sx]->p_rf = p_m[j]->p_rf;
        }

        if (p_sim->no_mark > 1 && !p_m[sx]->no_prop)
        {
            j = sx;
            if (p_m[AVG]->no_prop)
                j = AVG;
            else if (p_m[FEM]->no_prop)
                j = FEM;
            else if (p_m[MAL]->no_prop)
                j = MAL;
            else
                err_typ = ERA_MIP;

            p_m[sx]->no_prop = p_m[j]->no_prop;
            p_m[sx]->p_prop = p_m[j]->p_prop;
        }
        if (p_m[sx] && p_m[sx]->no_prop && p_sim->no_mark == 1)
        {
            err_typ = ERA_XIP;
            p_m[sx]->no_prop = 0;
        } 
    }
					/*now checks the sex-specific  	*/
                                        /*maps for consistency          */
    for (sx = FEM; sx < AVG; ++sx)
    {
        if (p_m[sx]->id != p_m[sx+1]->id)
            err_typ = ERD_TID;
        else if (p_m[sx]->no_xtrn != p_m[sx+1]->no_xtrn)
            err_typ = ERA_WXD;
        else if (p_m[sx]->no_prop != p_m[sx+1]->no_prop)
            err_typ = ERA_WIP;
    } 
    if (err_typ)
    {
        for_chrom (p_sim->chrom_no);

        return FALSE;
    }
    else 
        return TRUE;
}

void prnt_vmap(VMAP_PTR p_vm, int sx)
{
    int i;
    char some[10];

    if (p_vm->no_prop)
    {
        fprintf(stdout, "\n T%d ", p_vm->id);
        switch (sx)
        {
            case MAL:
                fprintf(stdout, "male ");
                break;
            case FEM:
                fprintf(stdout, "female ");
                break;
            default:
                break;
        }
        fprintf(stdout, "map interval proportions between markers:\n");
        for (i = 0; i < p_vm->no_prop; ++i)
        {
            if (!(i % 10))
                fprintf(stdout, "\n");
            sprintf(some, "%5.3f  ", *(p_vm->p_prop+i));   
            fprintf(stdout, " %s", &some[1]);
        }
        fprintf(stdout, "\n");
    } 

    fprintf(stdout, "\n T%d ", p_vm->id);
    switch (sx)
    {
        case MAL:
            fprintf(stdout, "male ");
            break;
        case FEM:
            fprintf(stdout, "female ");
            break;
        default:
            break;
    }
    fprintf(stdout, "map external distances (cM):\n");

    for (i = 0; i < p_vm->no_xtrn; ++i)
    {
        if (!(i % 10))
            fprintf(stdout, "\n");
        fprintf(stdout, " %4.1f  ", *(p_vm->p_dist+i));
    }
    fprintf(stdout, "\n\n");

    return;
}

void prnt_lods_freqs(SIM_PTR p_sim)
{
    char prnt_lin[81];

    if (!nxt_trait_vmap(FEM, p_sim))
        return;
    
    fprintf(stdout, " Trait allele frequencies:\n\n");
                                            	/*formats the freqs for */
                                            	/*the lods trait        */
    fmt_trait_freqs(prnt_lin, 0, nxt_trait_vmap(FEM, p_sim)->id);

    fprintf(stdout, "%s\n\n", prnt_lin);     	/* prints the line      */
    fflush(stdout);

    return;
}

void mak_chromo(PRNTMAP_PTR p_pmap)
{
    int i;
    PRNTLOC_PTR p_loc;

    if (NUnlnkd)
    {
        chromo[0][0] = 0;
        revchro[0][0] = 0;
    }

    p_loc = p_pmap->p_locus;

    for (i = NUnlnkd; i < NLoci; ++i)
    {
        if (p_loc->locus_typ == MD)
            chromo[0][i] = p_loc->id;
        else 
            chromo[0][i] = 0;

        revchro[0][chromo[0][i]] = i;

        p_loc = p_loc->p_next;
    }
    return;
}
 
void mak_xecomo(PRNTMAP_PTR p_pmap, int sx)
{
    int i;
    PRNTLOC_PTR p_loc;

    if (NUnlnkd)
        xecomo[sx][0][0] = .5;

    p_loc = p_pmap->p_locus;

    for (i = NUnlnkd; i < NLoci; ++i)
    {

        if (i < NLoci - 1) 
            xecomo[sx][0][i] = dist2rf(p_loc->p_next->loc_dist);
        else
            xecomo[sx][0][i] = .5;
     
        p_loc = p_loc->p_next;
    }
    return;
}

void alloc_lods_arrs(void)
{
    int sx;				/*allocates arrays		*/
                                   	/*Marek: 2 sexes, 		*/
                                        /*  # markers + 1 loci r.f.'s   */
                                        /*IntVal:  # trait pos + 1      */
                                        /*RecVal:  2 sexes,             */
                                        /*   # trait pos + 1 r.f.'s     */

    Marek = malloc(2 * sizeof(double *));
    alloc_chk(WHERE, Marek);
    RecVal = malloc(2 * sizeof(double *));
    alloc_chk(WHERE, RecVal);

    for (sx = FEM; sx <= MAL; ++sx)
    {
        Marek[sx] = malloc(NLoci * sizeof(double));
        alloc_chk(WHERE, Marek[sx]);
        RecVal[sx] = malloc((NTPos + 1) * sizeof(double));
        alloc_chk(WHERE, RecVal[sx]);
    }
    IntVal = malloc((NTPos + 1)* sizeof(int));
    alloc_chk(WHERE, IntVal);
    
    return;
}

void mak_intval(VMAP_PTR p_vm)
{
    int i, j;
                                     	/*sets loci for extern positions*/	
                                        /*plus one each end		*/ 
    for (i = 0; i <= p_vm->no_xtrn; ++i)
    {
        IntVal[i] = 0;
        IntVal[NTPos-i] = NMarks;
    }
    IntVal[NTPos] = 0;
                                    	/*and loci for interval         */
                                        /*proportions                   */
    for (j = 0; j < (NMarks - 1) * p_vm->no_prop; ++i, ++j)
        IntVal[i] = j / p_vm->no_prop + 1;

    return;
}

void mak_marek(PRNTMAP_PTR p_pmap, int sx)
{
    int i;
    PRNTLOC_PTR p_loc;
                                	/*makes Marek[][]     		*/ 
    p_loc = p_pmap->p_locus;
 
    for (i = 1; i < NMarks; ++i, p_loc = p_loc->p_next)
            Marek[sx][i] = dist2rf(p_loc->p_next->loc_dist);
     
    return;
}

void mak_recval(PRNTMAP_PTR p_pmap, VMAP_PTR p_vm, int sx)
{
    PRNTLOC_PTR p_cur; 
    int i, j, k;
                                     	/*set recomb fracts for each end*/	
    RecVal[sx][0] = .5;
    RecVal[sx][NTPos] = .5;
                                    	/*sets external recomb fracts   */
    j = p_vm->no_xtrn; 
    k = NTPos - p_vm->no_xtrn;
    for (i = 0; i < p_vm->no_xtrn; ++i, --j, ++k)
    {
        RecVal[sx][j] = *(p_vm->p_rf+i);
        RecVal[sx][k] = *(p_vm->p_rf+i);
    }
                                    	/*and recomb fracts between     */
                                        /*each marker pair              */
    p_cur = p_pmap->p_locus;
    while (p_cur && (p_cur = p_cur->p_next))
        for (j = 0; j < p_vm->no_prop; ++j)
        {
            RecVal[sx][i+1] = dist2rf( *(p_vm->p_prop+j) * 
                                        p_cur->loc_dist); 
            ++i;
        }

/*
    printf("\n NTPos: %d\n\n", NTPos);
    for (i = 0; i <= NTPos; ++i)
        printf(" IntVal[%d]: %d   RecVal[%d][%d]: %f\n", i, 
                         IntVal[i], sx, i, RecVal[sx][i]);
    printf("\n\n");  

*/

    return;
}

void mak_allefrq(PRNTMAP_PTR p_pmap, PRNTLOC_PTR p_ploc, int id)
{
    int j;
    LOCUS_PTR p_locus;

    if (!Qq)
    {
        Qq = malloc(NLoci * sizeof (double *));
        alloc_chk(WHERE, Qq);
        Qq[0] = malloc((MAX_ALLEL + 1) * (NLoci) * sizeof (double));   /**fix**/
        alloc_chk(WHERE, Qq[0]);
        for (j = 1; j <NLoci; ++j) 
            Qq[j] = Qq[j-1] + (MAX_ALLEL + 1);
      
        /*NAll = malloc(NLoci * sizeof (int));*/
        /*alloc_chk(WHERE, NAll);*/
    }

    if (p_ploc && p_ploc->locus_typ == MD)
    {
        p_locus = first_mark_locus_node;
        while ((p_locus->p_next 
                        && (p_locus->chrom_no < p_pmap->chrom_no)
                || (p_locus->chrom_no == p_pmap->chrom_no
                        && p_locus->id < p_ploc->id)))
            p_locus = p_locus->p_next;

        NAll[id] = p_locus->no_alleles;

        /*Qq[id] =  malloc ((NAll[id] + 1) * sizeof (double));*/
        /*alloc_chk(WHERE, Qq[id]);*/

        for (j = 0; j < NAll[id]; ++j)
            Qq[id][j + 1] = *(p_locus->p_freq_node->p_freq + j); 
    }
    else 
    {
        p_locus = first_trait_locus_node;
        while (p_locus->p_next && p_locus->id < id)
           p_locus = p_locus->p_next; 
       
        NAll[0] = p_locus->no_alleles;

        /*Qq[0] =  malloc ((NAll[0] + 1) * sizeof (double));*/
        /*alloc_chk(WHERE, Qq[0]);*/

        for (j = 0; j < NAll[0]; ++j)
            Qq[0][j + 1] = *(p_locus->p_freq_node->p_freq + j); 
    }
    return;
}

void mak_markdat(SIM_PTR p_sel)
{
    int ix, i, j;
    MARKDAT_PTR p_mdt;
  
    if (NObs == 0)
        return;

                                        /* finds marker data for this   */
    if (!(p_mdt = first_markdat_node))  /* chromosome                   */
        return;

    while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
        p_mdt = p_mdt->p_next;
    if (p_mdt->chrom_no != p_sel->chrom_no)
        return;

    ix = 0;

    for (i = 1; i <= NObs; ++i)
    {
        IObs[i] = *(p_mdt->p_list + ix);
        ++ix;
        for (j = 1; j <= NMarks; ++j)
        {
            Dat1[j][i] = *(p_mdt->p_list + ix);
            ++ix;
            Dat2[j][i] = *(p_mdt->p_list + ix);
            ++ix;
        }
    }
}

void mak_probgam(void)
{
    int c, i = 0;
    PROBGAM_PTR p_pg;

    p_pg = first_prob_gam_node;

    while (p_pg)
    {
        c = p_pg->component;
        NPr[c - 1] = p_pg->no_probs;

        for (i = 0; i < p_pg->no_probs; ++i)
        {
            Probands[c - 1][i]       = *(p_pg->p_list + 2 * i);
            Proband_gamete[c - 1][i] = *(p_pg->p_list + 2 * i + 1);
        }
    p_pg = p_pg->p_next;
    }
    return;
}

void mak_atz_inpt(void)		
{			
    int j;


    SelInt[0] = 1;			/* SelInt[]:  data[][]          */
                                        /* subscripts                   */

    if (parentOrdIn == UNDEF || parentOrdIn == PA1ST)
    {
	SelInt[1] = 2;			/* (father)			*/
	SelInt[2] = 3;			/* (mother)			*/
    } 
    else
    {
	SelInt[1] = 3;			/* (mother)			*/
	SelInt[2] = 2;			/* (father)			*/
    }
    SelInt[3] = 4;                      /* (gender)                     */

                                       /* allows for gender added to	*/
                                       /* pedigree 			*/ 
    for (j = 4; j < NInts + (SexIn == N ? 1 : 0); ++j)
    {
        if (first_tlist_node)
            if (j - 3 == first_tlist_node->id)
                SelInt[j] = 5;		/* 1 trait only        		*/
            else
                SelInt[j] = FALSE;
    }

    NColData = 6;                       /* # ints input + 1??           */

    return;
}

void prnt_markdat(SIM_PTR p_sel)
{
    int i, j;

    if (!NObs)
        return;

    if (p_sel->chrom_no)
        fprintf(stdout, " Marker data for chromosome %d:\n\n",
                        p_sel->chrom_no);
    else
        fprintf(stdout," Marker data:\n\n");

    for (i = 1; i <= NObs; ++i)
    {
        fprintf(stdout, " %d    ", IObs[i]);
        for (j = 1; j <= NMarks; ++j)
            fprintf(stdout, "%d %d  ", Dat1[j][i], Dat2[j][i]);
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n\n");
    fflush(stdout);

    return;
}

void prnt_probgam(void)
{
    int c, i = 0;

   
    fprintf(stdout," Proband gametes:\n\n");

    for (c = 1; c <= NC; ++c)
    {
        if (NPr[c - 1])
        {
            fprintf(stdout," Component %d:", c);
            for (i = 0; i < NPr[c - 1]; ++i)
            {
                fprintf (stdout, "  %d", Probands[c - 1][i]);
                fprintf (stdout, " %d", Proband_gamete[c - 1][i]);
            }
            fprintf (stdout, "\n");
        }
    }
    fprintf (stdout, "\n");
    fflush(stdout);

    return;
}

void prnt_atz_pars(void)
{
    if (TraitPheno)
        fprintf(stdout, " Trait data are phenotypes\n\n");
    else
        fprintf(stdout, " Trait data are genotypes\n\n");
 
    fprintf(stdout, " L-sampler to be used for setup\n\n"); 

    fprintf(stdout, " L-sampler probability set to %.3f\n\n", LSamplProb); 

    fprintf(stdout, " Number of MC iterations requested:  %d\n\n", NIt);

    if (Scan)
    {
        fprintf(stdout, " Sampling is to be by scan,");
        if (ScIt == 1)
            fprintf(stdout, " with scores estimated every scan\n\n");
        else
            fprintf(stdout, " with scores estimated every %d scans\n\n",ScIt);
    }
    else
    {
        fprintf(stdout, " Sampling is to be by single step,");
        if (ScIt == 1)
            fprintf(stdout, " with scores estimated every step\n\n");
        else
            fprintf(stdout, " with scores estimated every %d steps\n\n",ScIt);
    } 
    if (WIt)
        fprintf(stdout, " Progress to be checked every %d iterations\n\n",
                                        WIt);
    return;
}

