
#include "parsie.h"
#include "parsedefs.h"
#include "parsepars.h"
#include "parseproc.h"
#include "errspars.h"

/*
 * proc_pars() parses the parameter statements and stores the results
 * in intermediate tables.  Then proc_drop(), for the genedrop program,
 * or proc_poly(), for the polyEM programs, take this information and
 * prepare data for the specific MORGAN program.
 *							  MAJ Jul 1999 
 */

int proc_pars(int argc, char **argv, char *ppn)
{
 					/* gets name of calling program	*/
    for (pmx = prog; pmx->pnam[0] != NUL	
	             && strcmp(ppn, pmx->pnam); ++pmx)
	;
    set_defaults();

    if (!get_cmd_fnams(argc, argv))	/* gets command line file names	*/
        if (proc_err())
            return FALSE;
                                        /* process one file each loop	*/
    pfloop = Filtab;

    while (pfloop->type != NUL)
    {
        if (pfloop->fname && pfloop->parstat && !pfloop->done)
        {
            if (!open_par_fil())		/* opens input files	*/
	        if (proc_err())						
	            return FALSE;

            while (!pfloop->done)		/* processes 1 statement*/
            {					/* each loop		*/
             
                if (!get_tokens())		/* gets list of tokens  */
                { 
                    if (proc_err())
	                return FALSE;
                } 
	        else if (!get_aggr_key())	/* gets the statement	*/
	        {				/* type			*/
	            if (proc_err())
	                return FALSE;
	        } 
	        else if (!get_data_list()) 	/* gets the numeric or  */
	        {				/* string parameters   	*/
	            if (proc_err())
	                return FALSE;
	        } 
	        else if ((ppp->func)())		/* calls statement- 	*/
	        {                               /* specific functions   */
	            if (proc_err())
	                return FALSE;
	        }
            }
            if (!close_par_fil())               /* closes input files	*/
	        if (proc_err())						
	            return FALSE;

            fflush(stdout);
        }
        ++pfloop;				/* goes on to next file	*/ 	
    }  						/* type 		*/

    fflush (stdout);

    return TRUE;
}

/* -------- set_defaults:  sets default values to parameters ---------- */

void set_defaults(void)
{
    err_typ = 0;
                                        /* parameters for MORGAN progrms*/
					/* variables beginning with "-"	*/
                                        /* (underscore) are MORGAN	*/
                                        /* global variablesi            */
    ItMo = ITMO;
    ItEM = ITEM;
    ItWt = ITWT;
    Diff = DIFF;
    IPtEM= IEM;
    IEn = IEN;
    ICh = ICH;
    IOCor = IOC;
    ChkEig = N;
    ChkGnv = N;
    ChkGmt = N;
    ChkTrc = N;
    ChkEigC = N;

    parentOrdIn  = UNDEF;
    parentOrdOut = UNDEF;
    SexIn  = UNDEF;
    SexOp  = UNDEF;
    ObsIn  = UNDEF;
    ObsOp  = UNDEF;
    DiscardReals = UNDEF;
    SexOut = UNDEF;
    ObsOut = UNDEF;

    NIt  = NIT;
    WIt  = WIT;
    FWIt = FWIT;
    ScIt = 1;
    Scan = UNDEF;
    LSamplProb = 0.;
    TraitPheno = UNDEF;

    add_cor_flag = 0;
    res_cor_flag = 0;
    reset_fit_var(0);
    reset_fit_var(1);
    reset_fit_cov(0);
    reset_fit_cov(1);

    n_eigen = 0; 
    eigen_set = N;
    eigen_compute = N;

    MarkSeedSet = N;
    TraitSeedSet = N;
    SamplSeedSet = N;
    MarkSeedsOut = N;
    TraitSeedsOut = N;
    SamplSeedsOut = N;
    ChronOut = N;
    AllVarOut = N;

    max_npar = MAXNTOKEN;
    token = NULL;
    tok_typ  = NULL;
    j_par = NULL;
    u_par = NULL;
    d_par = NULL;
    realloc_arrs();

    max_tchar = MAXNTOKEN * NOMTOKLEN;
    token[0] = NULL;
    realloc_toks(0);

    return;
}

/* ---- get_cmd_fnams:  gets the file names from the command line -----	*/
/*                      checks the file types specified			*/
/*                      stores the names in Filtab[]                    */

int get_cmd_fnams(int argc, char **argv)
{
    int i;
    
    if (argc < 2)			/* tests for parameter file 	*/
    {				     	/* as first argument in command	*/
        err_typ = ERR_EIF;		/* line				*/
        return FALSE;
    }
    else	
    {
        pft = Filtab;
        pft->fname = alloc_copy(argv[1]);
    }
    for (i = 2; i < argc-1; i+=2)	/* identifies other files named	*/
    {					/* in command line		*/
        pft = get_filent(argv[i]);
        if (pft)
        {
            if (strchr(pft->pcods, pmx->pcod))	
            {	
                pft->command = Y;
       	        pft->fname = alloc_copy(argv[i+1]);
	    }
            else
	    {    			/* checks for file used by prog	*/
                err_typ = ERR_PCL;
                err_txt = pft->type;
            }                   
        }
        else 
        {
            err_typ = ERR_FNU;         	/* reports bad file type	*/
            err_txt = argv[i];
        }
    }
    if ((argc & 1) != 0)		/* tests for odd # of arguments	*/
    {
        i = 1;
        err_typ = ERR_FCL;
        sprintf(errnumb, COM_LIN_FORM);
        while (pft = get_filtyp(pmx->pcod))
        {
            if (i == 4)
            sprintf(&errnumb[strlen(errnumb)], "\n                      "); 
            sprintf(&errnumb[strlen(errnumb)], " [%s <%sfile>]", 
                                               pft->type, pft->type);
            ++i;
        }
        err_txt = errnumb;
    }
    if (err_typ)
        return FALSE;
    else
        return TRUE;
}

/* -------------- open_par_fil:  opens the input file ----------------- */
/*                   gets the first line of input			*/

int open_par_fil(void)
{
    char *ptok, *pstr;

					/* opens the file		*/
    errno = 0;
    if (!(pfile = fopen(pfloop->fname, pfloop->fop)))
    {
        err_typ = ERR_OPF;
        sprintf(errnumb, "%s \"%s\":  %s", pfloop->fdesc, 
                               pfloop->fname, strerror(errno));
        err_txt = errnumb; 
        pfloop->done = Y;
        return FALSE;
    }

    pstr = line_in;			/* tests for file not empty     */
    while (get_line() && !(ptok = strtok(pstr, WHITESPACE)))
	    ;
    if (err_typ)
        return FALSE;

    if (pfloop->done)
    {
        err_typ = ERR_SIF;
        err_txt = pfloop->fname;
        return FALSE;
    } 
    else
        *strchr(ptok, NUL) = ' ';	/* gets first token	 	*/

    return TRUE;
}

/* ------------- close_par_fil:  closes the input file ---------------- */

int close_par_fil(void)
{
					/* closes the file		*/
    errno = 0;
    if (pfile)
        if (fclose(pfile) == EOF)
        {
            err_typ = ERR_CLF;
            sprintf(errnumb, "%s \"%s\":  %s", pfloop->fdesc, 
                                   pfloop->fname, strerror(errno));
            err_txt = errnumb; 
        }
    pfile = NULL;

    if (err_typ)
        return FALSE;
    return TRUE; 
}

/* ---- chk_stor:  checks for enough storage allocated for arrays ----	*/

void chk_stor(int n_tok, int m_char)
{					/* n_tok:  # tokens stored	*/
    if (n_tok + 2 > max_npar)
    {
	max_npar *= 2;
	realloc_arrs();
    }					/* m_char:  string length	*/
    if (token[n_tok] - token[0] + m_char + 1 >= max_tchar)
    {
	max_tchar *= 2;
	realloc_toks(n_tok);
    }
    return;
}

/* -------- realloc_arrs:  allocates or reallocates storage ---------	*/
/*                          for arrays for input data 			*/

void realloc_arrs(void)
{
    token = realloc(token, max_npar * sizeof(char *));
    alloc_chk(WHERE, token);
    tok_typ = realloc(tok_typ, max_npar * sizeof(int));
    alloc_chk(WHERE, tok_typ);
    j_par = realloc(j_par, max_npar * sizeof(int));
    alloc_chk(WHERE, j_par);
    u_par = realloc(u_par, max_npar * sizeof(uint32));
    alloc_chk(WHERE, u_par);
    d_par = realloc(d_par, max_npar * sizeof(double));
    alloc_chk(WHERE, d_par);

    return;
}

/* --------- realloc_toks:  allocates or reallocates storage ---------	*/
/*                           for array of tokens 			*/

void realloc_toks(int n_tok)
{
    int i;
    char *ptmp;

    ptmp = token[0];
    token[0] = (char *) realloc(token[0], max_tchar * sizeof(char));
    alloc_chk(WHERE, token[0]);
    if (n_tok && token[0] != ptmp)
	for (i = 1; i <= n_tok; ++i)
	    token[i] = token[0] + (token[i] - ptmp);

    return;
}

/* ------- free_arrs:  frees all of proc_pars's malloc'd arrays -------	*/

void free_arrs(void)
{
    free((void *) token[0]);
    token[0] = NULL;
    free((void *) token);
    token = NULL;
    free((void *) tok_typ);
    tok_typ = NULL;
    free((void *) j_par);
    j_par = NULL;
    free((void *) u_par);
    u_par = NULL;
    free((void *) d_par);
    d_par = NULL;

    return;
}

/* ---------- get_line:  gets the next line of input data  ------------ */
/*                       replaces first "#" with "\0" (for comment)     */
/*                       returns pointer to line,                       */
/*                       unless eof or line contains "****"             */
/*                       as first keyword, then                         */
/*                       sets file pointer to NULL and returns NULL     */

char *get_line(void)
{
    char *pt, *ps;

    if (!fgets(line_in, MAXLINLEN + 2, pfile))
    {
        pfloop->done = Y;
	return NULL;
    } 
    else
    {
	ps = line_in;
        if (strlen(line_in) == MAXLINLEN + 1  && *(line_in + MAXLINLEN) != '\n')
        {
            err_typ = ERR_LTL;
            err_txt = line_in;
            pfloop->done = Y;
            return NULL;
        }
        if (pt = strchr(ps, COMMENT))
            *pt = NUL;
	if ((pt = strtok(ps, WHITESPACE)) != NULL)
	    if (!strncmp(ENDPAR, pt, 4))
	    {
                pfloop->done = Y;
		return NULL;
	    } 
            else
		*strchr(pt, NUL) = ' ';
	return line_in;
    }
}

/* -------------- get_tokens:  stores a list of tokens ---------------- */
/*                             returns the number found		        */

int get_tokens(void)
{
    char *ptok, *pstr;
    char chkverb[5];
    int end_statement;
    int n, i, m;

    pstr = line_in;	           	/* gets tokens for current line*/
    for (n = 0; (ptok = strtok(pstr, WHITESPACE)) != NULL; ++n)
    {
	m = (int) strlen(ptok) + 1;
	chk_stor(n, m);

	strncpy(token[n], ptok, m);
	token[n + 1] = token[n] + m;
	pstr = NULL;
    }
    token[n][0] = NUL;

    end_statement = FALSE;		/* gets the next line 		*/
    while (!pfloop->done && !end_statement)
    {
	if (get_line())
	{
	    pstr = line_in;		/* tests for end of statement   */
	    if ((ptok = strtok(pstr, WHITESPACE)) != NULL)
	    {
		if (isalpha((int)*ptok))
                {	
                    for (i = 0; i < 4 && *(ptok + i) != NUL; ++i)
                        chkverb[i] = tolower((int)*(ptok + i));
                    chkverb[i] = NUL;
		    for (i = 0; strcmp(verb[i], ""); ++i)
			if (!strncmp(verb[i], chkverb, 4))
			    end_statement = TRUE;
                }
		*strchr(ptok, NUL) = ' ';
		if (!end_statement)	/* gets tokens from the next lin*/
		{
		    pstr = line_in;
		    for (; (ptok = strtok(pstr, WHITESPACE)) != NULL; ++n)
		    {
			m = (int) strlen(ptok) + 1;
			chk_stor(n, m);
			strncpy(token[n], ptok, m);
			token[n + 1] = token[n] + m;
			pstr = NULL;
		    }
		    token[n][0] = NUL;
		}
	    }
	}
    }
    if (err_typ)
        return 0;

    return n;
}

/* -- get_aggr_key:  sets a pointer to the input statement prototype --	*/
/*                   returns key # from list				*/

int get_aggr_key(void)
{
    int i, j, n;

    aggre_key[0] = NUL;
    for (i = 0, n = 0; token[i][0] != NUL; ++i)
    {
	if (isalpha((int)token[i][0]))
	{
	    tok_typ[i] = KEYWD;
	    if (++n <= MAXNKEY)		/* gets aggregate key		*/
		strncat(aggre_key, token[i], 4);
	} 
        else
	    tok_typ[i] = PDAT;
    }
                                        /* converts to lower case       */
    for (i = 0; aggre_key[i] != NUL; ++i)
        aggre_key[i] = tolower((int)aggre_key[i]);

					/* searches for key in table	*/
    for (ppp = proto, j = 0; ppp->key[0] != NUL	
	 && strcmp(aggre_key, ppp->key) != 0; ++ppp, ++j)
	;
    if (ppp->key[0] != NUL)
	return key_ix = j + 1;
    else
    {
        err_typ = ERR_KEY;
	return key_ix = 0;
    }
}

/* ----- get_data_list:  converts tokens into integer and doubles -----	*/
/*                       returns FALSE if wrong number of numeric values*/

int get_data_list(void)
{
    int i, j, s;
    char c, *endp;

    ni_par = nk_par = nh_par = nr_par = nx_par = ns_par = 0;
    j = 0;

    for (i = 0; (c = ppp->par[i]) != NUL; ++i)
    {					/* processes all data (non- 	*/
                                        /* keywords) for the input      */
                                        /* statement			*/
	for (; token[j][0] != NUL && tok_typ[j] == KEYWD; ++j)
	    ;
	if (token[j][0] == NUL)
	{
	    err_typ = ERR_DAT;
	    return FALSE;
	}
	switch (c)
	{
	case 'i':			/* expects single integer value	*/
	    j_par[ni_par + nk_par] = atoi(token[j]);
	    ++ni_par;
	    break;
	case 'k':			/* expects list of int values	*/
	    for (; token[j][0] != NUL && tok_typ[j] == PDAT; ++j)
	    {
		j_par[ni_par + nk_par] = atoi(token[j]);
		++nk_par;
	    }
	    --j;
	    break;
        case 'h':			/* expects single hexadec value */
            u_par[nh_par] = (uint32) strtoul(token[j], &endp, 0);
            if (*endp != NUL)
                err_typ = ERR_HEX;
            ++nh_par;
            break;
	case 'r':			/* expects single real value	*/
	    d_par[nr_par + nx_par] = atof(token[j]);
	    ++nr_par;
	    break;
	case 'x':			/* expects list of real values	*/
	    for (; token[j][0] != NUL && tok_typ[j] == PDAT; ++j)
	    {
		d_par[nr_par + nx_par] = atof(token[j]);
		++nx_par;
	    }
	    --j;
	    break;
	case 's':			/* expects a string		*/
	    s = (int) strlen(token[j]) - 2;
            if (s < 1)
                err_typ = ERR_STR;
            if (token[j][0] != '\"' && token[j][0] != '\'')
                err_typ = ERR_STR;
            if (token[j][s+1] != '\"' && token[j][s+1] != '\'')
                err_typ = ERR_STR;
            if (!err_typ)
            {
	        strncpy(s_par, &token[j][1], s);
                s_par[s] = NUL;
            }
	    ++ns_par;
	    break;
	}
	++j;
    }

    if (err_typ)
        return FALSE;

    for (; token[j][0] != NUL; ++j)
	if (tok_typ[j] == PDAT)
	    err_typ = ERR_XTR;
    if (err_typ)
	return FALSE;
    else
	return TRUE;
}

/* ------ there is one function for each parameter statement type -----	 */
/*          they are named in a dispatch table in parsepars.h		 */

int not_impl(void)
{
    chk_stat();
    return err_typ;
}

int in_ped_fil(void)			/* input pedigree file	 	*/
{
    if (chk_stat())
        chk_fil(PED);

    return err_typ;
}

int in_pedsiz(void)			/* input pedsize statement	*/
{
    if (chk_stat())
    {
        chk_repeat();
	PedSize = j_par[0];
	if (PedSize < MIN_PED)
	    err_typ = ERR_PED;
	else
	if (PedSize > pmx->max_ped)
        {
	    err_typ = ERR_PEX;
            sprintf(errnumb, "%d", pmx->max_ped);
            err_txt = errnumb; 
        }
    }
    return err_typ;
}

int allow_pedsiz(void)			/* allow pedigree size statement*/
{
    if (chk_stat())
    {
        chk_repeat();
	allw_pedsiz = j_par[0];
	if (pmx->max_ped && allw_pedsiz > pmx->max_ped)
        {
            if (pmx->pcod == 'u')
		err_typ = ERR_ALU;
	    else
                err_typ = ERR_ALP;
            sprintf(errnumb, "%d", pmx->max_ped);
            err_txt = errnumb; 
        }
    }
    return err_typ;
}

int allow_cmpsiz(void)			/* allow component size statemnt*/
{
    if (chk_stat())
    {
        chk_repeat();
	allw_cmpsiz = j_par[0];
	if (pmx->max_cmp && allw_cmpsiz > pmx->max_cmp)
        {
            err_typ = ERR_ALC;
            sprintf(errnumb, "%d", pmx->max_cmp);
            err_txt = errnumb; 
        }
    }
    return err_typ;
}

int allow_obsiz(void)			/* allow observed individuals st*/
{
    if (chk_stat())
    {
        chk_repeat();
	allw_obsiz = j_par[0];
	if (pmx->max_obs && allw_obsiz > pmx->max_obs)
        {
            err_typ = ERR_ALO;
            sprintf(errnumb, "%d", pmx->max_obs);
            err_txt = errnumb; 
        }
    }
    return err_typ;
}

int allow_maxnam(void)			/* allow maximum name statement	*/
{
    if (chk_stat())
    {
        chk_repeat();
	allw_maxnam = j_par[0];
	if (allw_maxnam > MAX_NAMES)
        {
            err_typ = ERR_ALN;
            sprintf(errnumb, "%d", MAX_NAMES );
            err_txt = errnumb; 
        }
    }
    return err_typ;
}

int in_ped_int(void)			/* input pedigree record integer*/
{
    if (chk_stat())
    {
	chk_repeat();
	NInts = j_par[0];
	if (NInts < FMIN_INT || NInts > FMAX_INT)
	    err_typ = ERR_PF1;
        NReals = 0;
    }
    return err_typ;
}

int in_ped_int_rl(void)			/* input pedigr record int real */
{
    if (chk_stat())
    {
	chk_repeat();
	NInts = j_par[0];
	NReals = j_par[1];
	if (NInts < FMIN_INT || NInts > FMAX_INT)
	    err_typ = ERR_PF1;
	if (NReals < 0 || NReals > FMAX_REAL)
	    err_typ = ERR_PF2;
    }
    return err_typ;
}

int in_eig_fil(void)			/* input eigenvalue file	*/
{
    if (chk_stat())
        chk_fil(IGEN);

    return err_typ;
}

int in_peel_fil(void)			/* input peeling file		*/
{
    if (chk_stat())
        chk_fil(PEEL);

    return err_typ;
}

int in_seed_fil(void)			/* input seed file		*/
{
    if (chk_stat())
        chk_fil(SEED);

    return err_typ;
}

int out_ped_fil(void)			/* output ped file       	*/
{
    if (chk_stat())
        chk_fil(OPED);

    return err_typ;
}

int out_scor_fil(void)			/* output score file       	*/
{
    if (chk_stat())
        chk_fil(OSCORE);

    return err_typ;
}

int out_seed_fil(void)			/* output seed file       	*/
{
    if (chk_stat())
        chk_fil(OSEED);

    return err_typ;
}

int in_ord_fath(void)			/* input record father mother	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (parentOrdIn == MA1ST)
	    err_typ = ERR_ORD;
	parentOrdIn = PA1ST;
    }
    return err_typ;
}

int in_ord_moth(void)			/* input record mother father	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (parentOrdIn == PA1ST)
	    err_typ = ERR_ORD;
	parentOrdIn = MA1ST;
    }
    return err_typ;
}

int in_gend_abs(void)			/* input record gender absent	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (SexIn == Y)
	    err_typ = ERR_GND;
	SexIn = N;
    }
    return err_typ;
}

int in_gend_pres(void)			/* input record gender present	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (SexIn == N)
	    err_typ = ERR_GND;
	SexIn = Y;
    }
    return err_typ;
}

int in_obs_abs(void)			/* input record observed absent	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (ObsIn == Y)
	    err_typ = ERR_OPA;
	ObsIn = N;
    }
    return err_typ;
}

int in_obs_pres(void)			/* input record observed present*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (ObsIn == N)
	    err_typ = ERR_OPA;
	ObsIn = Y;
    }
    return err_typ;
}

int assum_all_obs(void)			/*assume all observed		*/
{
    if (chk_stat())
    {
        chk_repeat();
        ObsOp = ALL;
    }
    return err_typ;
}

int assign_gend(void)			/* assign gender          	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (SexOp == IGNORE)
	    err_typ = ERR_ASG;
        SexOp = ASSIGN;
    }
    return err_typ;
}

int ignor_gend(void)			/*  ignore gender	      	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (SexOp == ASSIGN)
	    err_typ = ERR_IGG;
	SexOp = IGNORE;
    }
    return err_typ;
}


int out_ord_fath(void)			/* output record father mother	*/
{
    if (chk_stat())
    {
	chk_repeat();
	if (parentOrdOut == MA1ST)
	    err_typ = ERR_ORD;
	parentOrdOut = PA1ST;
    }
    return err_typ;
}

int out_ord_moth(void)			/* output record mother father */
{
    if (chk_stat())
    {
	chk_repeat();
	if (parentOrdOut == PA1ST)
	    err_typ = ERR_ORD;
	parentOrdOut = MA1ST;
    }
    return err_typ;
}

int out_corr(void)			/* output correlations		*/
{
    if (chk_stat())
    {
	chk_repeat();
	IOCor = Y;
    }
    return err_typ;
}

int out_covar(void)			/* output covariances		*/
{
    if (chk_stat())
    {
	chk_repeat();
	IOCor = N;
    }
    return err_typ;
}

int lim_breed(void)			/* limit breeding iterations	*/
{
    if (chk_stat())
    {
	chk_repeat();
	ItMo = j_par[0];
    }
    return err_typ;
}

int lim_em_iter(void)			/* limit em iterations statement*/
{
    if (chk_stat())
    {
	chk_repeat();
	ItEM = j_par[0];
    }
    return err_typ;
}

int set_mc_iter(void)			/* set mc iteration  statement  */
{
    if (chk_stat())
    {
	chk_repeat();
	NIt = j_par[0];
    }
    return err_typ;
}

int set_scor_iter(void)			/* set scoring iter statement  */
{
    if (chk_stat())
    {
	chk_repeat();
	ScIt = j_par[0];
    }
    return err_typ;
}

int set_lsmpl_prob(void)
{
    if (chk_stat())
    {
	chk_repeat();
        if (d_par[0] < 0. || d_par[0] > 1.0)
            err_typ = ERA_BLS;

	LSamplProb = d_par[0];
    }
    return err_typ;
}

int set_trait_geno(void)
{
    if (chk_stat())
    {
	chk_repeat();
	if (TraitPheno == Y)
	    err_typ = ERR_STT;
        TraitPheno = N;
    }
    return err_typ;
}

int set_trait_pheno(void)
{
    if (chk_stat())
    {
	chk_repeat();
	if (TraitPheno == N)
	    err_typ = ERR_STT;
        TraitPheno = Y;
    }
    return err_typ;
}

int sampl_by_scan(void)
{
    if (chk_stat())
    {
	chk_repeat();
	if (Scan == N)
	    err_typ = ERR_SCN;
        Scan = Y;
    }
    return err_typ;
}
int sampl_by_step(void)
{
    if (chk_stat())
    {
	chk_repeat();
	if (Scan == Y)
	    err_typ = ERR_SCN;
        Scan = N;
    }
    return err_typ;
}

int out_scor_iter(void)			/* output scores MC iter statemnt*/
{
    if (chk_stat())
    {
	chk_repeat();
	FWIt = j_par[0];
    }
    return err_typ;
}

int out_space(void)			/* output spacing statement	*/
{
    if (chk_stat())
    {
	chk_repeat();
	ItWt = j_par[0];
    }
    return err_typ;
}

int use_full_em(void)			/* use full (regular) em	*/
{
    if (chk_stat())
    {
	chk_repeat();
        IPtEM = 0;
    }
    return err_typ;
}

int use_part_em(void)			/* use partitioned em		*/
{
    if (chk_stat())
    {
	chk_repeat();
        IPtEM = Y;
    }
    return err_typ;
}

int set_br_conv(void)			/* set breeding convergence	*/
{
    if (chk_stat())
    {
	chk_repeat();
	Diff = d_par[0];
    }
    return err_typ;
}

int fit_no_gen(void)			/* fit environmental model	*/
{
    if (chk_stat())
    {
	chk_repeat();
	IEn = 1;
    }
    return err_typ;
}

int sel_trait(void)			/* select trait statement	*/
{
    if (chk_stat())
    {
	if (chk_int_trait())
	{
	    if (!chk_sel_tr())
		 reset_sel_tr();
	    set_sel_tr(0);
	}
    }
    return err_typ;
}
int sel_tr_eff(void)			/* select trait... effect       */
{
    if (chk_stat())
    {
	if (chk_trait())
	{
	    if (!chk_sel_tr())
		 reset_sel_tr();
	    if (chk_ff(0))
		set_sel_tr(1);
	}
    }
    return err_typ;
}
int start_mean(void)			/* start trait mean statement	*/
{
    if (chk_stat())
    {
	if (chk_trait())
	{
	    if (!chk_start_ff(0))
		 reset_start_ff(0);
	    set_start_ff(0);
	}
    }
    return err_typ;
}
int start_effect(void)			/* start trait.. effect..       */
{
    if (chk_stat())
    {
	if (chk_trait())
	{
	    if (chk_ff(1))
	    {
		if (!chk_start_ff(1))
		     reset_start_ff(1);
		if (chk_levels_ff())
		    set_start_ff(1);
	    }
	}
    }
    return err_typ;
}
int start_add_var(void)			/* start additive variance..    */
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_var(0);
	if (chk_var())
	    set_start_var(0);

    }
    return err_typ;
}
int start_add_cor(void)			/* start additive correlation..	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_cov(0);
	if (chk_covar())
	    set_start_cor(0);
    }
    return err_typ;
}
int start_add_cov(void)			/* start additive covariance.. 	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_cov(0);
	if (chk_covar())
	    set_start_cov(0);
    }
    return err_typ;
}
int start_res_var(void)			/* start residual variance..    */
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_var(1);
	if (chk_var())
	    set_start_var(1);
    }
    return err_typ;
}
int start_res_cor(void)			/* start residual correlation..	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_cov(1);
	if (chk_covar())
	    set_start_cor(1);
    }
    return err_typ;
}
int start_res_cov(void)			/* start residual covariance.. 	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_start_cov(1);
	if (chk_covar())
	    set_start_cov(1);
    }
    return err_typ;
}
int fit_addi(void)			/* fit additive covariance.. 	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_fit_cov(0);
	if (chk_covar_i())
	    set_fit_cov(0);
    }
    return err_typ;
}
int fit_resi(void)			/* fit residual covariance.. 	*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_fit_cov(1);
	if (chk_covar_i())
	    set_fit_cov(1);
    }
    return err_typ;
}
int set_eigen(void)			/* set eigenvalues ..		*/
{
    if (chk_stat())
    {
	if (!chk_repeat())
	     reset_set_eigen();
	set_set_eigen();
    }
    return err_typ;
}
int comp_eigen(void)			/* compute eigenvalues statement*/
{
    if (chk_stat())
    {
        chk_repeat();
        ICh = 1;
    }
    return err_typ;
}
int check_eigen(void)			/* check eigenvalues		*/
{
    if (chk_stat())
    {
        chk_repeat();
	ChkEig = Y;
    }
    return err_typ;
}
int check_ginv(void)			/* check ginverse		*/
{
    if (chk_stat())
    {
	chk_repeat();
	ChkGnv = Y;
    }
    return err_typ;
}
int check_gmat(void)			/* check gmatrix		*/
{
    if (chk_stat())
    {
	chk_repeat();
	ChkGmt = Y;
    }
    return err_typ;
}

int check_trace(void)			/* check trace			*/
{
    if (chk_stat())
    {
	chk_repeat();
	ChkTrc = Y;
    }
    return err_typ;
}

int check_eig_cmp(void)			/* check eigenvalue computation	*/
{
    if (chk_stat())
    {
	chk_repeat();
	ChkEigC = Y;
    }
    return err_typ;
}


int check_mc_iter(void)			/* check MC iterations statemnt*/
{
    if (chk_stat())
    {
	chk_repeat();
	WIt = j_par[0];
    }
    return err_typ;
}

/* --- functions for "genedrop" and "lm_auto" parameter statements ---- */

int sim_mark(void)                      /* simulate markers             */
{
    SIM_PTR p_sim;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_sim(NCH, NT))
        {   
            if (!(p_sim = chk_sim_node(NCH))) /*tests for existing node  */ 
                p_sim = alloc_sim_node(NCH); 
            stor_sim_data(p_sim, NCH, NT);   /*stores data in node       */
        }
     }
     return err_typ;
}

int sim_cmark(void)                     /* sim markers on chromosome    */
{
    SIM_PTR p_sim;

    if (chk_stat())
        if (chk_sim(CH, NT))
        {   
            if (!(p_sim = chk_sim_node(CH)))  /*tests for existing node  */ 
                p_sim = alloc_sim_node(CH); 
            stor_sim_data(p_sim, CH, NT);    /*stores data in node       */
        }
    return err_typ;
}

int sim_mark_trait(void)                /* sim markers & traits         */
{
    SIM_PTR p_sim;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_sim(NCH, T))
        {   
            if (!(p_sim = chk_sim_node(NCH))) /*tests for existing node  */ 
                p_sim = alloc_sim_node(NCH); 
            stor_sim_data(p_sim, NCH, T);    /*stores data in node       */
        }
     }
     return err_typ;
}

int sim_cmark_trait(void)               /* sim markers & traits on chrom*/
{
    SIM_PTR p_sim;

    if (chk_stat())
        if (chk_sim(CH, T))
        {   
            if (!(p_sim = chk_sim_node(CH)))  /*tests for existing node  */ 
                p_sim = alloc_sim_node(CH); 
            stor_sim_data(p_sim, CH, T);     /*stores data in node       */
        }
    return err_typ;
}

int sim_unlinked(void)                  /* simulate unlinked traits     */
{
    UNLNKD_PTR p_unlnkd;
    int i;

    if (chk_stat())
    {
	if (!chk_repeat())              /* tests for existing trait list*/
            free_unlnkd();
        if (chk_unlnkd())
            for (i = 0; i < nk_par; i++)
            {
                p_unlnkd = alloc_unlnkd_node(j_par[i]);
                stor_unlnkd(j_par[i], p_unlnkd);
            }                           /* stores trait list            */
    }
    return err_typ;
}

int map_markr(void)                      /* map markers                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, AVG)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, R, MD);  /*stores data in node    */
        }
    }
    return err_typ;
}

int map_markd(void)                      /* map markers                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, AVG)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, D, MD); /*stores data in node      */
        }
    }
    return err_typ;
}

int map_markp(void)                      /* map markers                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, AVG))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, P, MD); /*stores data in node      */
        }
    }
    return err_typ;
}

int map_cmarkr(void)                     /* map markers on chromosome   */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, AVG))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, R, MD);  /*stores data in node      */
        }
    }
    return err_typ;
}

int map_cmarkd(void)                     /* map markers on chromosome   */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, AVG))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, D, MD);  /*stores data in node      */
        }
    }
    return err_typ;
}

int map_cmarkp(void)                     /* map markers on chromosome   */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, AVG)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, AVG); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, P, MD);  /*stores data in node      */
        }
    }
    return err_typ;
}

int map_traitd(void)                    /* map traits                   */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, AVG))
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, AVG); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, D, TD); /*stores data in node      */
        }
    }
    return err_typ;
}

int map_traitr(void)                    /* map traits                   */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, AVG))
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, AVG); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, R, TD); /*stores data in node      */
        }
    }
    return err_typ;
}

int map_ctraitd(void)                   /* map traits on chromosomes    */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, AVG)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, AVG); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, D, TD);  /*stores data in node      */
        }
    }
    return err_typ;
}

int map_ctraitr(void)                   /* map traits on chromosomes    */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, AVG))
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, AVG); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, R, TD);  /*stores data in node      */
        }
    }
    return err_typ;
}
                                           /* next 20 functions are     */
                                           /* for gender specific maps  */
                                           /* ...females first          */
int map_markr_f(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, FEM); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, R, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_markd_f(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, FEM); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, D, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_markp_f(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, FEM);
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, P, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkr_f(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, FEM); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, R, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkd_f(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, FEM); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, D, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkp_f(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, FEM); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, P, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_traitd_f(void)                    /* map traits                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, FEM); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, D, TD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_traitr_f(void)                    /* map traits                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, FEM); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, R, TD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_ctraitd_f(void)                   /* map traits on chromosomes  */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, FEM))
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, FEM); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, D, TD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_ctraitr_f(void)                   /* map traits on chromosomes  */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, FEM)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, FEM); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, R, TD);  /*stores data in node     */
        }
    }
    return err_typ;
}
                                           /* for gender specific maps  */
                                           /* ...males next             */

int map_markr_m(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, MAL)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, MAL); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, R, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_markd_m(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, MAL))
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, MAL); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, D, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_markp_m(void)                      /* map markers               */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, MD, MAL))
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, MD, MAL);
            alloc_dist(p_map, MD);
            stor_map_data(p_map, NCH, P, MD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkr_m(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, MAL))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, MAL); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, R, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkd_m(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, MAL))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, MAL); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, D, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_cmarkp_m(void)                     /* map markers on chromosome */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, P, MD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, MD, MAL))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, MD, MAL); 
            alloc_dist(p_map, MD);
            stor_map_data(p_map, CH, P, MD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_traitd_m(void)                    /* map traits                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, MAL))
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, MAL); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, D, TD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_traitr_m(void)                    /* map traits                 */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_map(NCH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(NCH, TD, MAL)) 
                free_dist(p_map);
            else
                p_map = alloc_map_node(NCH, TD, MAL); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, NCH, R, TD); /*stores data in node     */
        }
    }
    return err_typ;
}

int map_ctraitd_m(void)                   /* map traits on chromosomes  */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, D, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, MAL))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, MAL); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, D, TD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_ctraitr_m(void)                   /* map traits on chromosomes  */
{
    MAP_PTR p_map;

    if (chk_stat())
    {   
        if (chk_map(CH, R, TD))
        {                                /*     tests for existing node */ 
            if (p_map = chk_map_node(CH, TD, MAL))  
                free_dist(p_map);
            else
                p_map = alloc_map_node(CH, TD, MAL); 
            alloc_dist(p_map, TD);
            stor_map_data(p_map, CH, R, TD);  /*stores data in node     */
        }
    }
    return err_typ;
}

int map_ctr_intv_f(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, I, FEM)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, FEM)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, CH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_intv_m(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, I, MAL)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, MAL)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, CH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_intv(void)			/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, I, AVG)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, AVG)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, CH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_intv_f(void)			/* maps trait intervals 	*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(NCH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, I, FEM)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, FEM)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, NCH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_intv_m(void)			/* maps trait intervals         */
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(NCH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, I, MAL)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, MAL)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, NCH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_intv(void)			/* maps trait intervals         */
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(NCH, I))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, I, AVG)) 
                free_vdist(p_vmap, I);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, AVG)); 
            alloc_vdist(p_vmap, I);
            stor_vmap_data(p_vmap, NCH, I);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extd_f(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, D, FEM)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, FEM)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, CH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extr_f(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, R, FEM)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, FEM)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, CH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extd_m(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, D, MAL)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, MAL)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, CH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extr_m(void)		/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, R, MAL)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, MAL)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, CH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extd(void)			/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, D, AVG)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, AVG)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, CH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_ctr_extr(void)			/* maps trait intervals on chrom*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        if (chk_vmap(CH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(CH, R, AVG)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(CH, AVG)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, CH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extd_f(void)			/* maps trait external intrvals*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, D, FEM)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, FEM)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, NCH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extr_f(void)			/* maps trait intervals 	*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, R, FEM)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, FEM)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, NCH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extd_m(void)			/* maps trait external intrvals*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, D, MAL)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, MAL)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, NCH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extr_m(void)			/* maps trait intervals         */
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, R, MAL)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, MAL)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, NCH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extd(void)			/* maps trait external intrvals*/
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, D))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, D, AVG)) 
                free_vdist(p_vmap, D);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, AVG)); 
            alloc_vdist(p_vmap, D);
            stor_vmap_data(p_vmap, NCH, D);    /*stores data in node    */
        }
    }
    return err_typ;
}

int map_tr_extr(void)			/* maps trait intervals         */
{
    VMAP_PTR p_vmap;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_vmap(NCH, R))
        {                             	/*     tests for existing node */ 
            if (p_vmap = chk_vmap_node(NCH, R, AVG)) 
                free_vdist(p_vmap, R);
            else
                init_vmap_lists(p_vmap = alloc_vmap_node(NCH, AVG)); 
            alloc_vdist(p_vmap, R);
            stor_vmap_data(p_vmap, NCH, R);    /*stores data in node    */
        }
    }
    return err_typ;
}

int set_mark_frq(void)                  /* set marker frequencies       */
{
    LOCUS_PTR p_locus;
    FREQ_PTR p_fnode;
    int i;

    if (chk_stat())
    {
        if (chk_locus(NCH, MD))
        {
            p_fnode = alloc_frq();
            stor_frq(p_fnode);
            for (i = 0;  i < nk_par; i++)
            {
                if (p_locus = chk_locus_node(j_par[i], NCH, MD))
                    free_freq(p_locus);
                else
                    p_locus = alloc_locus_node(j_par[i], NCH, MD);
                stor_locus_data(j_par[i], p_locus, p_fnode, NCH, MD);
            }
        }
    }
    return err_typ;
}

int set_cmark_frq(void)                 /* set chrom/marker frequencies */
{
    LOCUS_PTR p_locus;
    FREQ_PTR p_fnode;
    int i;

    if (chk_stat())
    {
        if (chk_locus(CH, MD))
        {
            p_fnode = alloc_frq();
            stor_frq(p_fnode);
            for (i = 0;  i < nk_par; i++)
            {
                if (p_locus = chk_locus_node(j_par[i+1], CH, MD))
                    free_freq(p_locus);
                else
                    p_locus = alloc_locus_node(j_par[i+1], CH, MD);
                stor_locus_data(j_par[i+1], p_locus, p_fnode, CH, MD);
            }
        }
    }
    return err_typ;
}

int set_trait_frq(void)                 /* set trait frequencies        */
{
    LOCUS_PTR p_locus;
    FREQ_PTR p_fnode;
    int i;

    if (chk_stat())
    {
        if (chk_locus(NCH, TD))
        {
            p_fnode = alloc_frq();
            stor_frq(p_fnode);
            for (i = 0;  i < nk_par; i++)
            {
                if (p_locus = chk_locus_node(j_par[i], NCH, TD))
                    free_freq(p_locus);
                else
                    p_locus = alloc_locus_node(j_par[i], NCH, TD);
                stor_locus_data(j_par[i], p_locus, p_fnode, NCH, TD);
            }
        }
    }
    return err_typ;
}

int set_geno_mean(void)                 /* set trait genotype means     */
{                                       /* (must be 3 means)            */
    MEANS_PTR p_means;
    int i;

    if (chk_stat())
    {
        if (chk_means())
            for (i = 0; i < nk_par; i++)
            {
                if (!(p_means = chk_means_node(j_par[i])))
                    p_means = alloc_means_node(j_par[i]);
                stor_mean_data(j_par[i], p_means);
            }
    }
    return err_typ;
} 

int set_addi_var(void)                  /* set additive variances       */
{
    VARI_PTR p_vari;
    int i;

    if (chk_stat())
    {
        if (chk_vari())
            for (i = 0; i < nk_par; i++)
            {
                if (!(p_vari = chk_vari_node(j_par[i], AV)))
                    p_vari = alloc_vari_node(j_par[i], AV);
                stor_vari_data(i, p_vari);
            }
    }
    return err_typ;
}

int set_resi_var(void)                  /* set residual variances       */
{
    VARI_PTR p_vari;
    int i;

    if (chk_stat())
    {
        if (chk_vari())
            for (i = 0; i < nk_par; i++)
            {
                if (!(p_vari = chk_vari_node(j_par[i], RV)))
                    p_vari = alloc_vari_node(j_par[i], RV);
                stor_vari_data(i, p_vari);
            }
    }
    return err_typ;
}

int set_mark_seeds(void)                 /* set marker seeds            */
{
    if (chk_stat())
    {
        chk_repeat();
        chk_seeds();
        MarkSeedSet = Y;
        MarkSeedC = u_par[0];
        MarkSeedT = u_par[1];
    }
    return err_typ;
}

int set_trait_seeds(void)                /* set trait seeds             */
{
    if (chk_stat())
    {
        chk_repeat();
        chk_seeds();
        TraitSeedSet = Y;
        TraitSeedC = u_par[0];
        TraitSeedT = u_par[1];
    }
    return err_typ;
}

int set_sampl_seeds(void)                /* set sampler seeds           */
{
    if (chk_stat())
    {
        chk_repeat();
        chk_seeds();
        SamplSeedSet = Y;
        SamplSeedC = u_par[0];
        SamplSeedT = u_par[1];
    }
    return err_typ;
}

int out_mark_seeds(void)                 /* output marker seeds only    */
{
    if (chk_stat())
    {
        chk_repeat();
        if (TraitSeedsOut)
        {
            TraitSeedsOut = N;
            err_typ = ERD_OTS;
        }
        MarkSeedsOut = Y; 
    }
    return err_typ;
}

int out_trait_seeds(void)                /* output trait seeds only     */
{
    if (chk_stat())
    {
        chk_repeat();
        if (MarkSeedsOut)
        {
            MarkSeedsOut = N;
            err_typ = ERD_OMS;
        }
        TraitSeedsOut = Y; 
    }
    return err_typ;
}

int out_all(void)			/* output pedigree variables all*/
{
    if (chk_stat())
    {
        chk_repeat();
        AllVarOut = Y;
    }
    return err_typ;
}

int set_mark_dat(void)                  /* set marker data              */
{
    MARKDAT_PTR p_mdnode;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_mrkdat(NCH))
        {                               /* tests for existing node      */
            if (!(p_mdnode = chk_mrkdat_node(NCH)))
                p_mdnode = alloc_mrkdat_node(NCH); 
            else
                free_mrkdat_list(p_mdnode);
            alloc_mrkdat_list(p_mdnode);
            stor_mrkdat(p_mdnode, NCH);  /*stores data in node & in list */
        }
    }
    return err_typ;
}

int set_cmark_dat(void)                 /* set marker data for chromosom*/
{
    MARKDAT_PTR p_mdnode;

    if (chk_stat())
        if (chk_mrkdat(CH))
        {                               /* tests for existing node      */
            if (!(p_mdnode = chk_mrkdat_node(CH)))
                p_mdnode = alloc_mrkdat_node(CH); 
            else
                free_mrkdat_list(p_mdnode);
            alloc_mrkdat_list(p_mdnode);
            stor_mrkdat(p_mdnode, CH);   /*stores data in node & in list */
        }
    return err_typ;
}

int set_prbgam(void)                    /* set proband gametes          */
{
    PROBGAM_PTR p_pgnode;

    if (chk_stat())
    {   
        chk_repeat();
        if (chk_prbgam(NCMP))
        {                               /* tests for existing node      */
            if (!(p_pgnode = chk_prbgam_node(NCMP)))
                p_pgnode = alloc_prbgam_node(NCMP); 
            else
                free_prbgam_list(p_pgnode);
            alloc_prbgam_list(p_pgnode);
            stor_prbgam(p_pgnode, NCMP);/* stores data in node & in list*/
        }                                   
     }
     return err_typ;
}

int set_cmp_prbgam(void)                /* set compnt.. proband gametes */
{
    PROBGAM_PTR p_pgnode;

    if (chk_stat())
    {   
        if (chk_prbgam(CMP))
        {                               /* tests for existing node      */
            if (!(p_pgnode = chk_prbgam_node(CMP)))
                p_pgnode = alloc_prbgam_node(CMP); 
            else
                free_prbgam_list(p_pgnode);
            alloc_prbgam_list(p_pgnode);
            stor_prbgam(p_pgnode, CMP); /* stores data in node & in list*/
        }                          
     }
     return err_typ;
}

int titl_top(void)                      /* title                        */
{
    return err_typ;
}

int label_mark(void)                    /* label markers                */
{
    return err_typ;
}

int label_trait(void)                   /* label traits                 */
{
    return err_typ;
}

int out_chron(void)                    /* output pedigree chronological */
{
    if(chk_stat())
        chk_repeat();
    ChronOut = Y;

    return err_typ;
}

/* ------------ end of functions for parameter statements ------------- */

/* ------------ functions used by these functions follow -------------- */

/* -- check fncs:  check contents of parameter statements for errors -- */

int chk_stat(void)
{					/* increments counter for this  */
    ++ppp->count;			/* statement type		*/

    if (strchr(ppp->pcods, pmx->pcod))	/* checks for this statement 	*/
	return TRUE;			/* implemented for program	*/
    else
    {
	err_typ = ERR_NOT;
        err_txt = pmx->pnam;
	return FALSE;
    }
 }
int chk_repeat(void)			/* test for redundant statement	*/
{
    if (ppp->count > 1)
    {
	err_typ = ERR_RED;
	return FALSE;
    } else
	return TRUE;
}

void chk_fil(char * f_typ)             /* tests for file name already   */
                                       /* given.                        */
{                                      /* Stores file name              */
    pft = get_filent(f_typ);

    if (!pft)
       return;
 
    if (pft->command)
    {
        err_typ = ERR_CFO;
        sprintf(errnumb, "%s", pft->fdesc);
        err_txt = errnumb; 
    }
    else
    {
        if (!chk_repeat())
            free_fnam();
        pft->fname = alloc_copy(s_par);
        pft->done = N;
    }
    pft->param = Y;

    return;
}

int chk_trait(void)			/* tests for qualitative trait  */
                                        /* no out of range              */
{
    if (j_par[0] < 1 || j_par[0] > FMAX_REAL)
    {
	err_typ = ERR_TRO;
	return FALSE;
    } else
	return j_par[0];
}


int chk_int_trait(void)			/* tests for quantitative trait */
                                        /* no out of range              */
{
    if (j_par[0] < 1 || j_par[0] > FMAX_INT)
    {
	err_typ = ERR_TRO;
	return FALSE;
    } else
	return j_par[0];
}

int chk_sel_tr(void)			/* tests for repeated trait #	*/
{
    if (SelTrait[j_par[0] - 1])
    {
	err_typ = ERR_TRR;
	return FALSE;
    } else
	return j_par[0];
}
int chk_ff(int m)			/* tests for fixed effect no.  	*/
{					/* out of range			*/
    int k;

    if (!m)				/* select			*/
    {
	for (k = 1; k <= nk_par; ++k)
	    if (j_par[k] < 1 || j_par[k] > FMAX_FEF)
	    {
		err_typ = ERR_FFO;
		return FALSE;
	    }
	return TRUE;
    } else				/* start		        */
    if (j_par[1] < 1 || j_par[1] > FMAX_FEF)
    {
	err_typ = ERR_FFO;
	return FALSE;
    } else
	return TRUE;
}
int chk_start_ff(int m)			/* tests for mean or fixed      */
{					/* effect levels already given  */
                                        /* for trait                    */
 
    if (!m)				/* mean		 		  */
    {
	if (nstartff[j_par[0] - 1][0] == 1)
	{
	    err_typ = ERR_MNR;
	    return FALSE;
	} else
	    return TRUE;
    } else				/* fixed effect			*/
    {
	if (nstartff[j_par[0] - 1][j_par[1]] != 0)
	{
	    err_typ = ERR_FFR;
	    return FALSE;
	} else
	    return TRUE;
    }
}
int chk_levels_ff(void)			/* tests for too many starting	*/
{					/* levels for fixed effect	*/
    if (nx_par > LEVELS)
    {
	err_typ = ERR_LVM;
	return FALSE;
    } else
	return TRUE;
}
int chk_var(void)			/* tests for too many starting	*/
{					/* values for variances 	*/
    if (nx_par > pmx->max_trait)
    {
	err_typ = ERR_VAM;
	return FALSE;
    } else
	return TRUE;
}
int chk_covar(void)			/* tests for too many starting	*/
{					/* values for covariances	*/
    if (nx_par > pmx->max_trait * (pmx->max_trait - 1) / 2)
    {
	err_typ = ERR_COM;
	return FALSE;
    } else
	return TRUE;
}
int chk_covar_i(void)			/* tests for too many fit flags	*/
{					/* for covariances		*/
    if (nk_par > pmx->max_trait * (pmx->max_trait - 1) / 2)
    {
	err_typ = ERR_CFM;
	return FALSE;
    } else
	return TRUE;
}
int chk_sim(int c, int t)               /* tests for chromosome no.,    */
{                                       /* # markers, # traits, out of  */
    int ix = 0;                         /* range                        */

    if (CH == c)
    {
        if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
        {
            err_typ = ERD_CHO;
            return FALSE;
        }
        ix++;
    }
    if (j_par[ix] < 1)
    {
        err_typ = ERD_MK0;
        return FALSE;
    }
    if (j_par[ix] > pmx->max_mark)
    {
        err_typ = ERD_MKM;
        return FALSE;
    }   
    ix++;

    if (T == t)
    {
        if (j_par[ix] < 0 )
        {
            err_typ = ERD_TR0;
            return FALSE;
        }
        if (j_par[ix] > pmx->max_trait)
        {
            err_typ = ERD_TRM;
            return FALSE;
        }
    }
    return TRUE;
}
			
SIM * chk_sim_node(int c)               /* tests for statement already  */
{                                       /* given for this chromosome no */
    SIM_PTR p_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    if (p_node = first_sim_node)
    {
        while (p_node->p_next && p_node->chrom_no < n)
            p_node = p_node->p_next;

        if (p_node->chrom_no == n)
        {
            if (n == 0)
                err_typ = ERD_RSM;
            else
                err_typ = ERD_RCS;
            return p_node;
        }
    }
    return NULL;
}

int chk_unlnkd(void)                    /* tests for # traits and trait */
{                                       /* no's out of range            */
    int i;

    if (nk_par > pmx->max_trait)
    {
        err_typ = ERD_UTM; return FALSE;
    }
    for (i = 0; i < nk_par; i++)
        if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
        {
            err_typ = ERD_TRO;
            return FALSE;
        }
    return TRUE;
}

int chk_map(int c, int d, int m)        /* tests for chromosome no.,    */
{                                       /* # markers, # traits or #     */
                                        /* distances out of range       */
                                        /* also tests distances or      */
    int i, ix = 0;                      /* recomb fractions within range*/
    int err_ret = FALSE;

    if (CH == c)                         /* checks chromosome no.        */
    {
        if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
        {
            err_typ = ERD_CHO;
            err_ret = TRUE; 
        }
        ix++;
    }
    if (err_ret)
        return FALSE;

    if (m == MD)  			/* checks map marker statement  */
        switch (d)
        {
            case R:
                if (nx_par >= pmx->max_mark)
                {
                    err_typ = ERD_MRM;
                    err_ret = TRUE;
                }
                break;
            case D:
                if (nx_par >= pmx->max_mark)
                {
                    err_typ = ERD_MDM;
                    err_ret = TRUE;
                }
                break;
            case P:
                if (nx_par > pmx->max_mark)
                {
                    err_typ = ERD_MPM;
                    err_ret = TRUE;
                }
                break;
            default:
                break;
        }

    if (err_ret)
        return FALSE;

    if (m == TD) 			/* checks map trait statement   */
    {
        if (nk_par != 2 * nx_par)
        {
            err_typ = ERD_TMM;
            err_ret = TRUE;
        }
        else if (nx_par > pmx->max_trait)
        {
            if (R == d)
                err_typ = ERD_MRM;
            else if (D == d)
                err_typ = ERD_MDM;
            err_ret = TRUE;
        }
        if (!err_ret)
            for (i = 0; i < nk_par/2; i++)
                if (j_par[ix+i] < 1 || j_par[ix+i] > pmx->max_trait)
                {
                    err_typ = ERD_TRO;
                    err_ret = TRUE;
                }
        if (!err_ret)
            for (i = 0; i < nk_par/2; i++)
                if (j_par[ix+nk_par/2+i] < 0 || 
                         j_par[ix+nk_par/2+i] > pmx->max_mark)
                {
                    err_typ = ERD_MKO;
                    err_ret = TRUE;
                }
    }
    if (err_ret)
        return FALSE;

    for (i = 0; i < nx_par; i++)        /* checks fractions, distances  */
    {                                   /* or positions                 */
        switch (d)
        {                               /* tests for recomb fract >= 0. */
            case R:                     /* and < .5                     */
                if (d_par[i] < 0. || d_par[i] >= .5)
                {
                    if (m == MD)
                        err_typ = ERD_MFO;
                    else
                        err_typ = ERD_TFO;
                    err_ret = TRUE;
                }
                break;
                                        /* tests for distance >= 0.     */
            case D:                     /* and <= 500. cM               */
                if (d_par[i] < 0. || d_par[i] > 500.)
                {
                    if (m == MD)
                        err_typ = ERD_MDO;
                    else
                        err_typ = ERD_TDO;
                    err_ret = TRUE;
                }
                break;
                                        /* tests for position >= -500.  */
                                        /* and <= 500. cM               */
            case P:                     /* also tests for increasing pos*/
                if (d_par[i] < -500. || d_par[i] > 500.)
                {
                    err_typ = ERD_MPO;
                    err_ret = TRUE;
                }
                else if (i && d_par[i] < d_par[i-1])
                {
                    err_typ = ERD_MPD;
                    err_ret = TRUE;
                }
                break;

            default:
                break;
        }
    }
    if (err_ret)
        return FALSE;
    else
        return TRUE;
}

int chk_vmap(int c, int d)              /* tests for chromosome no. or  */
{                                       /* trait no. out of range       */  
                                        /* also tests interval propor-  */
                                        /* tions, external distances or */
                                        /* recomb fractions within range*/
    int i, ix = 0;                      /* and in increasing order      */
    int err_ret = FALSE;

    if (CH == c)                         /* checks chromosome no.        */
    {
        if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
        {
            err_typ = ERD_CHO;
            return FALSE; 
        }
        ix++;
    }

    if (j_par[ix] < 1 || j_par[ix] > pmx->max_trait)
    {
        err_typ = ERD_TRO;
        return FALSE; 
    }

    for (i = 0; i < nx_par; i++)        /* checks interval proportions, */
    {                                   /* fractions or distances       */
        switch (d)
        {                               /* tests for proportions > 0.   */
            case I:                     /* and < 1.                     */

                if (d_par[i] <= 0. || d_par[i] >= 1.)
                {
                    err_typ = ERA_TIO;
                    err_ret = TRUE;
                }                       /* checks for increasing order  */
                else if (i && d_par[i] <= d_par[i-1])
                {
                    err_typ = ERA_TID;
                    err_ret = TRUE;
                }
                break;
                                        /* tests for recomb fract > 0.  */
            case R:                     /* and < .5                     */

                if (d_par[i] <= 0. || d_par[i] >= .5)
                {
                    err_typ = ERA_TFO;
                    err_ret = TRUE;
                }                       /* checks for increasing order  */
                else if (i && d_par[i] <= d_par[i-1])
                {
                    err_typ = ERA_TFD;
                    err_ret = TRUE;
                }
                break;
                                        /* tests for distance > 0.      */
            case D:                     /* and <= 500. cM               */

                if (d_par[i] <= 0.)
                {
                    err_typ = ERA_TDO;
                    err_ret = TRUE;
                }      
                else if (d_par[i] > 500.)
                {
                    err_typ = ERA_TDR;
                    err_ret = TRUE;
                }                       /* checks for increasing order  */
                else if (i && d_par[i] <= d_par[i-1])
                {
                    err_typ = ERA_TDD;
                    err_ret = TRUE;
                }
                break;

            default:
                break;
        }
    }
    if (err_ret)
        return FALSE;
    else
        return TRUE;
}

MAP * chk_map_node(int c, int m, int sx) /* tests for statement already */
{                                        /* given for this chromosome no*/
    MAP_PTR p_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    if (m == MD)  
        p_node = first_mark_map_node[sx];
    else
        p_node = first_trait_map_node[sx];

    if (p_node)
    {
        while (p_node->p_next && p_node->chrom_no < n)
            p_node = p_node->p_next;

        if (p_node->chrom_no == n)
        {
            if (n == 0)
                err_typ = ERD_RMP;
            else
                err_typ = ERD_RCS;
            return p_node;
        }
    }
    return NULL;
}

VMAP * chk_vmap_node(int c, int d, int sx)
                                        /* tests for statement already 	*/
                                        /* given for this chromosome no */
 					/* and for this type interval   */
{					/* (internal or external)       */
    VMAP_PTR p_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    p_node = first_trait_vmap_node[sx];

    if (p_node)
    {
        while (p_node->p_next && p_node->chrom_no < n)
            p_node = p_node->p_next;

        if (p_node->chrom_no == n)
        {
            if (I == d  &&  p_node->p_prop  ||
                D == d  &&  p_node->p_dist  ||
                R == d  &&  p_node->p_dist)  
            { 
                if (n == 0)
                    err_typ = ERD_RMP;
                else
                    err_typ = ERD_RCS;

            }
            return p_node;
        }
    }
    return NULL;
}

int chk_locus(int c, int md)            /* tests for chromosome no.,    */
                                        /* marker no's, trait no's,     */
                                        /* # allele frequencies within  */
                                        /* range.  checks sum of allele */
{                                       /* frequencies                  */
    int i, ix = 0;
    double sum = 0;
    double norm;

    if (CH == c)
    {
        if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
        {
            err_typ = ERD_CHO;
            return FALSE;
        }
        ix++;
    }
    for (i = 0; i < nk_par; i++)
    {
        if (MD == md)
        {
            if (j_par[ix] < 1 || j_par[ix] > pmx->max_mark)
            {
                err_typ = ERD_MKO;
                return FALSE;
            }
        }
        else
            if (j_par[ix] < 1 || j_par[ix] > pmx->max_trait)
            {
                 err_typ = ERD_TRO;
                 return FALSE;
            }
        ix++;
    }
    if (MD == md)
    {
        if (nx_par < MIN_ALLEL || nx_par > pmx->max_allele)
        {
            err_typ = ERD_MLO;
            return FALSE;
        }
    }
    else
    {
        if (nx_par != 2)
        {
            err_typ = ERD_TLO;
            return FALSE;
         }
    }
    for (i = 0; i < nx_par; i++)
    {
        if (d_par[i] < 0. || d_par[i] > 1.0)
        {
            err_typ = ERD_AFO;
            return FALSE;
        }
        sum += d_par[i];
    }
    if (sum < 1. - ALLEF_TOL || sum > 1. + ALLEF_TOL)
        err_typ = ERD_SMA;

    norm = 1.0/sum;
    for (i = 0; i < nx_par; i++)
        d_par[i] *= norm; 


    return TRUE;
} 
LOCUS * chk_locus_node(int no, int c, int md)  /* checks for allele     */
                                               /* frequencies already   */
{                                              /* given for this locus  */
    LOCUS_PTR p_node;
    int n = 0;
						
    if (MD == md)                       /* marker allele frequencies    */
    {
        if (CH == c)
            n = j_par[0];
        p_node = first_mark_locus_node;
    }
    else                               /* trait allele frequencies      */
    {
        n = NO_CH_NO;
        p_node = first_trait_locus_node;
    }
    if (p_node)
    {
        while (p_node->p_next 
                   && (p_node->chrom_no < n 
                   || (p_node->chrom_no == n && p_node->id < no)))
            p_node = p_node->p_next;
        if (p_node->chrom_no == n && p_node->id == no)
        {
            if (MD == md)
                err_typ = ERD_MAF;     /* overrides previous allele frqs*/
            else                       /* for marker                    */
                err_typ = ERD_TAF;     /* or for trait                  */
            sprintf(errnumb, "%d", p_node->id);
            err_txt = errnumb;
            return p_node;
        }
    }
    return NULL;                       /* no match for id               */
}

int chk_means(void)                    /* tests for trait # in range,   */
                                       /* and  for # means = 3          */
{
    int i;

    for (i = 0; i < nk_par; i++)
        if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
        {
            err_typ = ERD_TRO;
            return FALSE;
        }
    if (nx_par != 3)
    {
        err_typ = ERD_TGM;
        return FALSE;
    }
    return TRUE;
}

MEANS * chk_means_node(int no)         /* tests for trait genotype means*/
                                       /* already given for this trait  */
{
    MEANS_PTR p_node;

    p_node = first_means_node;
    if (p_node)
    {
        while (p_node->p_next && p_node->id < no)
            p_node = p_node->p_next;
        if (p_node->id == no) 
        {
            err_typ = ERD_TGR;
            return p_node;
        }
    }
    return NULL;
}

int chk_vari(void)                     /* tests for trait # in range,   */
{                    
    int i;

    for (i = 0; i < nk_par; i++)
        if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
        {
            err_typ = ERD_TRO;
            return FALSE;
        }
    if (d_par[0] < 0.)
    {
        err_typ = ERD_VMN;
        return FALSE;
    }
    return TRUE;
}

VARI * chk_vari_node(int no, int v)    /* tests for variance already    */
                                       /* given for this trait          */
{
    VARI_PTR p_node;

    if (AV == v)
        p_node = first_addi_node;
    else
        p_node = first_resi_node;

    if (p_node)
    {
        while (p_node->p_next && p_node->id < no)
            p_node = p_node->p_next;

        if (p_node->id == no)
        {
            if (AV == v)
                err_typ = ERD_AVR;
            else
                err_typ = ERD_RVR;
            return p_node;
        }
    }
    return NULL;
}

int chk_mrkdat(int c)                   /* tests for chromosome no.,    */
{                                       /* out of range                 */
                                        /* checks for at least 3? data  */
    int i, ix = 0;                      /* and for data not negative    */
    int err_ret = FALSE;

    if (CH == c)
    {
        if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
        {
            err_typ = ERD_CHO;
            err_ret = TRUE;
        }
        ix++;
    }
    if (nk_par < 3)
    {
        err_typ = ERA_MDL;
        err_ret = TRUE;
    }
    else
        for ( i = 0; i < nk_par ; ++i)
            if (j_par[ix + i] < 0)
            {
                err_typ = ERA_BMD;
                sprintf(errnumb, "%d", j_par[ix + i]);
                err_txt = errnumb;
                err_ret = TRUE;
            }
    if (err_ret)
        return FALSE;
    else
        return TRUE;
}
			
MARKDAT * chk_mrkdat_node(int c)        /* tests for statement already  */
{                                       /* given for this chromosome no */
    MARKDAT_PTR p_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    if (p_node = first_markdat_node)
    {
        while (p_node->p_next && p_node->chrom_no < n)
            p_node = p_node->p_next;

        if (p_node->chrom_no == n)
        {
            if (n == 0)
                err_typ = ERA_RMD;
            else
                err_typ = ERD_RCS;
            return p_node;
        }
    }
    return NULL;
}

int chk_prbgam(int cmp)                 /* tests for component no.      */
{                                       /* out of range                 */
                                        /* checks name and gamete       */
    int i, ix = 0;                 
    int err_ret = FALSE;

    if (CMP == cmp)
    {
        if (j_par[ix] < 1 || j_par[ix] > pmx->max_n_cmp)
        {
            err_typ = ERA_CMP;
            sprintf(errnumb, "%d", j_par[ix]);
            err_txt = errnumb;
            err_ret = TRUE;
        }
        ix++;
    }
    if (err_ret)
       return FALSE;

    if ((nk_par) % 2 != 0)
    {
        err_typ = ERA_PGP;
        err_ret = TRUE;
    }
    else if (nk_par < 4)
    {
        err_typ = ERA_PGL;
        err_ret = TRUE;
    }
    else if (nk_par / 2 >= G6)
    {
        err_typ = ERA_PGX;
        sprintf(errnumb, "%d", G6 - 1);
        err_txt = errnumb;
        err_ret = TRUE;
    }
    if (err_ret)
        return FALSE;
    
    for (i = 0; i < nk_par / 2; ++i)
    {
        if (j_par[ix] <= 0)
        {
            err_typ = ERA_BPG;
            sprintf(errnumb, "%d", j_par[ix]);
            err_txt = errnumb;
            err_ret = TRUE;
        }
        else if (j_par[ix + 1] != 0 && j_par[ix + 1] != 1)
        {
            err_typ = ERA_BGM;
            sprintf(errnumb, "%d", j_par[ix+1]);
            err_txt = errnumb;
            err_ret = TRUE;
        }
        ix += 2;
    }
    if (err_ret)
        return FALSE;

    return TRUE;
}
			
PROBGAM * chk_prbgam_node(int cmp)      /* tests for statement already  */
{                                       /* given for this component     */
    PROBGAM_PTR p_node;

    int cp_no = 0;
    int ix = 0;

    if (CMP == cmp)
        cp_no = j_par[ix];

    if (p_node = first_prob_gam_node)
    {
        while (p_node->p_next && p_node->component < cp_no)
            p_node = p_node->p_next;

        if (p_node->component == cp_no)
        {
            if (cp_no == 0)
                err_typ = ERD_RMP;
            else if (cp_no != 0)
                err_typ = ERD_RCS;

            return p_node;
        }
    }
    return NULL;
}

void chk_seeds(void)
{
    if (u_par[0] % 2 == 0)
    {
        u_par[0] |= 1;
        err_typ = ERD_SEE;
    }
    if (u_par[1] == 0)
    {
        u_par[1] = 1;
        err_typ = ERD_SEE;
    }
    return;
}
       
/* -------------- free fncs:  frees list nodes or arrays -------------- */

void free_unlnkd(void)                   /* frees current list of       */
                                         /* unlinked traits             */
{
    UNLNKD_PTR p_node, p_nxtn;

    p_node = first_unlnkd_node;

    while (p_node != NULL)
    {
        p_nxtn = p_node->p_next;
        free(p_node);
        p_node = p_nxtn;
    }
    first_unlnkd_node = NULL;
   
    return;
}

void free_dist(MAP_PTR p_node)
{
    if (p_node->p_dist != NULL)
    {
        free (p_node->p_dist);
        p_node->p_dist = NULL;
    }
    return;
}

void free_vdist(VMAP_PTR p_node, int d)
{
    if (d == I)
    {
        if (p_node->p_prop != NULL)
        {
            free (p_node->p_prop);
            p_node->p_prop = NULL;
            p_node->no_prop = 0;
        }
    }
    else
    {
        if (p_node->p_dist != NULL)
        {
            free (p_node->p_dist);
            p_node->p_dist = NULL;
            free (p_node->p_rf);
            p_node->p_rf = NULL;
            p_node->no_xtrn = 0;
        }
    }
    return;
}

void free_freq(LOCUS_PTR p_locus)        /* if there are no loci for the*/
                                         /* frequencies, free list      */
{                                
    if ((p_locus->p_freq_node->no_loci -= 1) == 0)  /* decrement & test */
    {                                               /* # loci           */
        free (p_locus->p_freq_node->p_freq);
        free (p_locus->p_freq_node);
    }
    p_locus->p_freq_node = NULL;
    p_locus->no_alleles = 0;

    return;
}

void free_mrkdat_list(MARKDAT_PTR p_node)
{
    if (p_node->p_list != NULL)
    {
        free (p_node->p_list);
        p_node->p_list = NULL;
    }
    return;
}

void free_prbgam_list(PROBGAM_PTR p_node)
{
    if (p_node->p_list != NULL)
    {
        free (p_node->p_list);
        p_node->p_list = NULL;
    }
    return;
}

/* ------------- alloc fncs:  malloc list nodes or arrays ------------- */

SIM * alloc_sim_node(int c)              /* allocates the sim node      */
{                                        /* inserts in list in numeric  */
    SIM_PTR new_node, cur_node;          /* order                       */

    int n = 0;

    if (CH == c)
        n = j_par[0];

    new_node = malloc(sizeof(SIM));
    alloc_chk(WHERE, new_node);

    if (!first_sim_node || first_sim_node->chrom_no > n)
    {
        new_node->p_next = first_sim_node;     /* puts at start of list */
        first_sim_node = new_node;             /* (may be only node)    */
    }
    else
    {
        cur_node = first_sim_node;             /* inserts node in list  */
                                               /* or adds to end of list*/
        while (cur_node->p_next && cur_node->p_next->chrom_no < n)
            cur_node = cur_node->p_next;
        
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

UNLNKD * alloc_unlnkd_node(int no)       /* allocates an unlinked trait */
{                                        /* node                        */
    UNLNKD_PTR new_node, cur_node;

    new_node = malloc(sizeof(UNLNKD));
    alloc_chk(WHERE, new_node);

    if (!first_unlnkd_node || first_unlnkd_node->id > no)
    {
        new_node->p_next = first_unlnkd_node;   /* puts at start of list*/
        first_unlnkd_node = new_node;
    }
    else
    {
        cur_node = first_unlnkd_node;           /* puts after start of  */
                                                /* list                 */
        while (cur_node->p_next && cur_node->p_next->id < no)
            cur_node = cur_node->p_next;

        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

MAP * alloc_map_node(int c, int m, int sx)
{                                        /* allocates the map node      */
                                         /* inserts in list in numeric  */
    MAP_PTR new_node, cur_node;          /* order                       */
    MAP_PTR * first_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    new_node = malloc(sizeof(MAP));
    alloc_chk(WHERE, new_node);

    if (m == MD)  
        first_node = &first_mark_map_node[sx];
    else
        first_node = &first_trait_map_node[sx];

    if (!*first_node || (*first_node)->chrom_no > n)
    {
        new_node->p_next = *first_node;    /*     puts at start of list */
        *first_node = new_node;            /*     (may be only node)    */
    }
    else
    {
        cur_node = *first_node;            /*     inserts node in list  */
                                           /*     or adds to end of list*/
        while (cur_node->p_next && cur_node->p_next->chrom_no < n)
            cur_node = cur_node->p_next;
        
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

VMAP * alloc_vmap_node(int c, int sx)
{                                        /* allocates the map node      */
                                         /* inserts in list in numeric  */
    VMAP_PTR new_node, cur_node;         /* order                       */
    VMAP_PTR * first_node;

    int n = 0;

    if (CH == c)
        n = j_par[0];

    new_node = malloc(sizeof(VMAP));
    alloc_chk(WHERE, new_node);

    first_node = &first_trait_vmap_node[sx];

    if (!*first_node || (*first_node)->chrom_no > n)
    {
        new_node->p_next = *first_node;    /*     puts at start of list */
        *first_node = new_node;            /*     (may be only node)    */
    }
    else
    {
        cur_node = *first_node;            /*     inserts node in list  */
                                           /*     or adds to end of list*/
        while (cur_node->p_next && cur_node->p_next->chrom_no < n)
            cur_node = cur_node->p_next;
        
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

void init_vmap_lists(VMAP_PTR p_v)
{
    p_v->no_prop = 0;
    p_v->p_prop = NULL;
    p_v->no_xtrn = 0;
    p_v->p_dist = NULL;
    p_v->p_rf = NULL;

    return;
} 

void alloc_dist(MAP_PTR p_node, int m)            /* allocates the      */
{                                                 /* distance array     */

    if (m == MD)  
        p_node->p_dist = malloc((nx_par + 1) * sizeof(DIST));
    else
        p_node->p_dist = malloc(nx_par * sizeof(DIST));
    alloc_chk(WHERE, p_node->p_dist);
    
    return;
}

void alloc_vdist(VMAP_PTR p_node, int d)/* allocates the interval   	*/
{                                       /* proportions or the external  */
                                        /* distances or recomb fracts   */
                                        /* array                        */
    if (d == I)
    {
        p_node->p_prop = malloc(nx_par * sizeof(double));
        alloc_chk(WHERE, p_node->p_prop);
        p_node->no_prop = 0;
    }
    else
    {
        p_node->p_dist = malloc(nx_par * sizeof(double));
        alloc_chk(WHERE, p_node->p_dist);
        p_node->no_xtrn = 0;
        p_node->p_rf = malloc(nx_par * sizeof(double));
        alloc_chk(WHERE, p_node->p_rf);
    }
 
    return;
}

FREQ_PTR alloc_frq(void)                          /* allocates frequency*/
{                                                 /* header and array   */
    return (alloc_frqx(nx_par));
}

FREQ_PTR alloc_dflt_frq(void)                     /* allocates frequency*/
{                                                 /* header and array   */
    return (alloc_frqx(2));                       /* for default values */
}



FREQ_PTR alloc_frqx(int n_alleles)                /* allocates frequency*/
{                                                 /* header and array   */
    FREQ_PTR p_f;

    p_f = malloc(sizeof(FREQ));
    alloc_chk(WHERE, p_f);
    p_f->p_freq = malloc(n_alleles * sizeof(double));
    alloc_chk(WHERE, p_f->p_freq);
    return p_f;
}

LOCUS_PTR alloc_locus_node(int no, int c, int md)
{
    int cno = 0;

    if (MD == md)                       /* marker allele frequencies    */
    {
        if (CH == c)
            cno = j_par[0];
    }
    else                               /* trait allele frequencies      */
        cno = NO_CH_NO;

    return (alloc_locus_nodex(cno, no, md));
}

LOCUS_PTR alloc_dflt_locus_node(int cno, int no, int md)
{
    return (alloc_locus_nodex(cno, no, md));
}

LOCUS * alloc_locus_nodex(int chrom, int no, int md)
{
    LOCUS_PTR new_node, cur_node;
    LOCUS_PTR * first_node;

    if (MD == md)                    
        first_node = &first_mark_locus_node;
    else                       
        first_node = &first_trait_locus_node;

    new_node = malloc(sizeof(LOCUS));
    alloc_chk(WHERE, new_node);

    if (!*first_node 
            || (*first_node)->chrom_no > chrom 
            || (*first_node)->chrom_no == chrom && (*first_node)->id > no)
    {
        new_node->p_next = *first_node;    /* puts at start of list     */
        *first_node = new_node;
    }
    else
    {
        cur_node = *first_node;

        while (cur_node->p_next 
                        && (cur_node->p_next->chrom_no < chrom 
                            || (cur_node->p_next->chrom_no == chrom 
                                && cur_node->p_next->id < no)))
            cur_node = cur_node->p_next;
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

MEANS * alloc_means_node(int no)         /* allocates the means node    */
                                         /* inserts in list in numeric  */
                                         /* order                       */
{
    MEANS_PTR new_node, cur_node;

    new_node = malloc(sizeof(MEANS));
    alloc_chk(WHERE, new_node);

    if (!first_means_node || first_means_node->id > no)
    {
        new_node->p_next = first_means_node;
        first_means_node = new_node;
    }
    else
    {
        cur_node = first_means_node;
        while (cur_node->p_next && cur_node->p_next->id < no)
            cur_node = cur_node->p_next;
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

VARI * alloc_vari_node(int no, int v)    /* allocates the variance node */
                                         /* inserts in list in numeric  */
                                         /* trait order                 */
{
    VARI_PTR new_node, cur_node;
    VARI_PTR * first_node;

    if (AV == v)
        first_node = &first_addi_node;
    else
        first_node = &first_resi_node;

    new_node = malloc(sizeof(VARI));
    alloc_chk(WHERE, new_node);

    if (!*first_node || (*first_node)->id > no)
    {
        new_node->p_next = *first_node;
        *first_node = new_node;
    }
    else
    {
        cur_node = *first_node;
        while (cur_node->p_next && cur_node->p_next->id < no)
            cur_node = cur_node->p_next;
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

MARKDAT * alloc_mrkdat_node(int c)       /* allocates marker data node  */
{                                        /* inserts in list in numeric  */
    MARKDAT_PTR new_node, cur_node;      /* order                       */

    int n = 0;

    if (CH == c)
        n = j_par[0];

    new_node = malloc(sizeof(MARKDAT));
    alloc_chk(WHERE, new_node);

    if (!first_markdat_node || first_markdat_node->chrom_no > n)
    {
        new_node->p_next = first_markdat_node; /* puts at start of list */
        first_markdat_node = new_node;         /* (may be only node)    */
    }
    else
    {
        cur_node = first_markdat_node;         /* inserts node in list  */
                                               /* or adds to end of list*/
        while (cur_node->p_next && cur_node->p_next->chrom_no < n)
            cur_node = cur_node->p_next;
        
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

void alloc_mrkdat_list(MARKDAT_PTR p_node)        /* allocates the      */
{                                                 /* marker data array  */

    p_node->p_list = malloc((nk_par) * sizeof(int));
    alloc_chk(WHERE, p_node->p_list);
    
    return;
}

PROBGAM * alloc_prbgam_node(int cmp)     /* allocates marker data node*/
{                                        /* inserts in list in numeric  */
    PROBGAM_PTR new_node, cur_node;      /* order                       */

    int cp_no = 0;

    if (CMP == cmp)
        cp_no = j_par[0]; 

    new_node = malloc(sizeof(PROBGAM));
    alloc_chk(WHERE, new_node);

    if (!first_prob_gam_node || first_prob_gam_node->component > cp_no)
    {
        new_node->p_next = first_prob_gam_node; /* puts at start of list */
        first_prob_gam_node = new_node;         /* (may be only node)    */
    }
    else
    {
        cur_node = first_prob_gam_node;        /* inserts node in list  */
                                               /* or adds to end of list*/
        while (cur_node->p_next && cur_node->p_next->component < cp_no)
            cur_node = cur_node->p_next;
        
        new_node->p_next = cur_node->p_next;
        cur_node->p_next = new_node;
    }
    return new_node;
}

void alloc_prbgam_list(PROBGAM_PTR p_node)        /* allocates the      */
{                                                 /* marker data array  */

    p_node->p_list = malloc((nk_par) * sizeof(int));
    alloc_chk(WHERE, p_node->p_list);
    
    return;
}

/* ------------ stor fncs:  stores data in nodes or arrays ------------ */

void stor_sim_data(SIM_PTR p_node, int c, int t) /* stores the chromosom*/
{                                                /* no., # markers and  */
    int ix = 0;                                  /* # traits in the sim */
                                                 /* node                */
    if (CH == c)
        p_node->chrom_no = j_par[ix++];
    else
        p_node->chrom_no = 0;

    p_node->no_mark = j_par[ix++];

    if  (T == t)
        p_node->no_traits = j_par[ix++];
    else
        p_node->no_traits = 0;

    return;
}

void stor_unlnkd(int no, UNLNKD_PTR p_node)    /* stores the trait no.  */
{                                              /* in the unlnkd node    */
    p_node-> id = no;

    return;
}

void stor_map_data(MAP_PTR p_node, int c, int d, int m) 
					/* stores the chromosome no. &	*/
					/* # loci in the map node.	*/
					/* stores an array of trait or	*/
					/* marker id's, preceding marker*/
					/* id's, probabilities & 	*/
					/* distances in the distance	*/
					/* structure.			*/
					/* for traits, sorts the 	*/
					/* array.			*/
{                          
    int i;              
    int ix = 0, jx = 0;
    DIST_PTR p_r;  
                 
    if (CH == c)
        p_node->chrom_no = j_par[ix++];
    else
        p_node->chrom_no = 0;

    if (m == MD)  
    {
        if (P != d)
            p_node->no_loci = nx_par + 1;
        else 
            p_node->no_loci = nx_par;
    }
    else
        p_node->no_loci = nx_par;
 
    p_r = p_node->p_dist;
    for (i = 0; i < p_node->no_loci; ++i)
    {
        if (m == MD)  
        {
            p_r->id = i + 1;
            p_r->mark_id = i;
        }
        else
        {
            p_r->id = j_par[ix];
            p_r->mark_id = j_par[ix+p_node->no_loci];
            ++ix;
        }
    
        if (i == 0 && m == MD) 
            {
                p_r->dist = NIL;
                p_r->prob = .5;
            }
        else
        {
            switch (d)
            {    
                case R:
                    p_r->dist = -.5 * log(1. - 2. * d_par[jx]) * 100.;
                    p_r->prob = d_par[jx];
                    break;
              
                case D:
                    p_r->dist = d_par[jx];
                    p_r->prob = .5 * (1. - exp(-2. * d_par[jx] / 100.));
                    break;

                case P:
                    p_r->dist = d_par[jx+1] - d_par[jx];
                    p_r->prob = .5 * 
                      (1. - exp(-2. * (d_par[jx+1] - d_par[jx]) / 100.));
                    break;

                default:
                    break;
            }
            ++jx;
        }
        ++p_r;
    }
    if (m == TD)
        qsort(p_node->p_dist, p_node->no_loci, sizeof(DIST), 
                               compar_mkid_d);

    return;
}

void stor_vmap_data(VMAP_PTR p_node, int c, int d) 
					/* stores the chromosome no. &	*/
					/* # of interval or external    */
                                        /* positions in the map node.	*/
					/* stores the array of          */
                                        /* proportions or external      */
                                        /* distances or recombination	*/
                                        /* fractions in a distance node */
{                          
    int i, ix = 0;              
    double * p_d;  
                 
    if (CH == c)
        p_node->chrom_no = j_par[ix++];
    else
        p_node->chrom_no = 0;

    p_node->id = j_par[ix];

    switch (d)
    {
        case I:				/* stores interval proprotions	*/
            p_node->no_prop = nx_par;
            p_d = p_node->p_prop;
            break;

        case D:   			/* stores external distances	*/
            p_node->no_xtrn = nx_par;	/* and converts to recomb fract */
            p_d = p_node->p_dist;
            for (i = 0; i < nx_par; ++i)
               p_node->p_rf[i] = dist2rf(d_par[i]);
            break;
        case R:				/* stores external recomb fracts*/
            p_node->no_xtrn = nx_par;	/* and converts to distances	*/
            p_d = p_node->p_rf;
            for (i = 0; i < nx_par; ++i)
               p_node->p_dist[i] = rf2dist(d_par[i]);
            break;
        default:
            break;
    }
    for (i = 0; i < nx_par; ++i)
        p_d[i] = d_par[i];

    return;
}

int compar_mkid_d(const void * p1, const void * p2)  
{                                       /* called by qsort for distance	*/
					/* array			*/
					/* orders on marker id & dist	*/
    DIST_PTR pd1;
    DIST_PTR pd2;

    pd1 = (DIST_PTR const) p1;
    pd2 = (DIST_PTR const) p2;

    if (pd1->mark_id < pd2->mark_id
            || (pd1->mark_id == pd2->mark_id && pd1->dist < pd2->dist))
        return -1;

    else    if (pd1->mark_id == pd2->mark_id && pd1->dist == pd2->dist)
        return 0;
    else 
        return 1;
}

void stor_frq(FREQ_PTR p_f)             /* stores the number of loci    */
                                        /* for these alleles in the     */
                                        /* header and stores an array   */
{                                       /* of alleles                   */
    int i;       
    
    p_f->no_loci = nk_par;
   
    for (i = 0; i < nx_par;  ++i)
        *(p_f->p_freq + i) = d_par[i];

    return;
}

void stor_dflt_frq(FREQ_PTR p_f)        /* stores the default allele    */
                                        /* frequencies                  */
{        
    p_f->no_loci = 1;
    *(p_f->p_freq) = .5;
    *(p_f->p_freq + 1) = .5;

    return;
}

void stor_locus_data(int no, LOCUS_PTR p_l, FREQ_PTR p_f, int c, int md)
{                                       /* stores data in the locus node*/
    if (MD == md)
    {
        if (CH == c)
            p_l->chrom_no = j_par[0];      
        else 
            p_l->chrom_no = 0;
    }
    else
        p_l->chrom_no = NO_CH_NO;

    p_l->id = no;
    p_l->no_alleles = nx_par;
    p_l->p_freq_node = p_f;

    return;
}

void stor_dflt_locus_data(int chro, int no, LOCUS_PTR p_l, FREQ_PTR p_f)
{
    p_l->chrom_no = chro;		/*stores locus node data	*/
    p_l->id = no;			/*for default allele frequencies*/
    p_l->no_alleles = 2;
    p_l->p_freq_node = p_f;

    return;
}

void stor_mean_data(int no, MEANS_PTR p_node)
{
    int i;

    p_node->id = no;
    for (i=0; i<3; i++)
        p_node->mean[i] = d_par[i];

    return;
}

void stor_dflt_mean_data(int no, MEANS_PTR p_node)
{
    int i;

    p_node->id = no;
    for (i=0; i<3; i++)
        p_node->mean[i] = 0.;

    return;
}

void stor_vari_data(int i, VARI_PTR p_node)
{
    p_node->id = j_par[i];
    p_node->vari = d_par[0];
    
    return;
}

void stor_dflt_vari_data(int no, VARI_PTR p_node)
{
    p_node->id = no;
    p_node->vari = 0.;

    return;
}

void stor_mrkdat(MARKDAT_PTR p_md, int c)
{                                       /* stores data in the marker    */
                                        /* data node.  Also stores the  */
                                        /* array of marker data         */
    int i, ix = 0;

    if (CH == c)
        p_md->chrom_no = j_par[ix++];      
    else
        p_md->chrom_no = 0;

    p_md->no_ints = nk_par;

    for (i = 0 ; i < nk_par;  ++i)
        *(p_md->p_list + i) = j_par[ix + i];

    return;
}

void stor_prbgam(PROBGAM_PTR p_pg, int cmp)
{                                       /* stores data in the marker    */
                                        /* data node.  Also stores the  */
                                        /* array of marker data         */
    int i, ix = 0;

    if (CMP == cmp)
        p_pg->component = j_par[ix++];
    else
        p_pg->component = 1;

    p_pg->no_probs = nk_par / 2;

    for (i = 0; i < nk_par;  ++i)
        *(p_pg->p_list + i) = j_par[ix + i];

    return;
}

/* ----------- copy fnc:  copies string to malloc'd storage ----------- */

char * alloc_copy(char * text)		/* first gets storage, then 	*/
{					/* copies			*/
    char * pstrg;
    int s;

    s = (int) strlen(text) + 1; 
    pstrg = malloc(s * sizeof(char *));
    alloc_chk(WHERE, pstrg);
    strncpy(pstrg, text, s);

    return pstrg;
}

/* -------------- reset fncs:  reinitialize data to 0's --------------- */

void reset_sel_tr(void)			/* resets selection/nonselection*/
{					/* of the trait and fixed effcts*/
    int k;		

    SelTrait[j_par[0] - 1] = FALSE;
    for (k = 0; k < FMAX_FEF; ++k)
	seltrff[j_par[0] - 1][k] = FALSE;
    return;
}
void reset_start_ff(int m)		/* resets starting value(s) for	*/
{					/* the mean or fixed effects for*/
    int k;				/* the trait			*/

    if (!m)				/* mean				*/
    {
	nstartff[j_par[0] - 1][0] = 0;
	startff[j_par[0] - 1][0][0] = 0.0;
	return;
    } else				/* fixed effect			*/
    {
	nstartff[j_par[0] - 1][j_par[1]] = 0;
	for (k = 0; k < LEVELS; ++k)
	    startff[j_par[0] - 1][j_par[1]][k] = 0.0;
	return;
    }
}
void reset_start_var(int m)		/* resets the starting value(s)	*/
{					/* for variances		*/
    int j;
   
    for (j = 1; j <= pmx->max_trait; ++j)
	if (!m)				/* additive			*/
	    SIGSQ(j, j) = 0.0;
	else				/* residual			*/
	    TAUSQ(j, j) = 0.0;
    if (!m)
	naddvar = 0;
    else
	nresvar = 0;
    return;
}
void reset_start_cov(int m)		/* resets the starting value(s)	*/
{					/* for covariances		*/
    int row, col; 

    for (row = 2; row <= pmx->max_trait; ++row)
        for (col = 1; col < row; ++col)
	    if (!m)			/* additive			*/
		SIGSQ(row, col) = 0.0;
	    else			/* residual			*/
		TAUSQ(row, col) = 0.0;
    if (!m)
	naddcov = 0;
    else
	nrescov = 0;
    return;
}
void reset_fit_var(int m)		/* resets the fit/nofit flags	*/
{					/* for variances (= 1)		*/
    int j; 

    for (j = 1; j <= pmx->max_trait; ++j)
	if (!m)				/* additive			*/
       	    SIG_MODL(j, j) = 1;
 	else				/* residual			*/
	    TAU_MODL(j, j) = 1;
}
void reset_fit_cov(int m)		/* resets the fit/nofit flags	*/
{					/* for covariances		*/
    int row, col; 

    for (row = 2; row <= pmx->max_trait; ++row)
        for (col = 1; col < row; ++col)
	    if (!m)			/* additive			*/
		SIG_MODL(row, col) = 1;
	    else			/* residual			*/
		TAU_MODL(row, col) = 1;
    if (!m)
	nfitadd = 0;
    else
	nfitres = 0;
    return;
}
void free_fnam(void)			/* frees the string containing	*/
{					/* the file name		*/
    free((void *) pft->fname);
    pft->fname = NULL;
    return;
}
void reset_set_eigen(void)		/* frees the eigenvalue array	*/
{
    free((void *) peigen);
    peigen = NULL;
    return;
}
/* ------ set fncs:  store input parameters into data structures ------ */

void set_sel_tr(int m)			/* identifies the trait and fixd*/
{					/* effects, if any, selected    */
    int k;

    SelTrait[j_par[0] - 1] = TRUE;

    if (m)
	for (k = 1; k <= nk_par; ++k)
	{
	    if (seltrff[j_par[0] - 1][j_par[k]])
		err_typ = ERR_FFD;
	    else
		seltrff[j_par[0] - 1][j_par[k]] = TRUE;
	}
    return;
}
void set_start_ff(int m)		/* sets the starting value(s)   */
{					/* for the mean or fixed effects*/
    int k;				/* for the trait	        */

    if (!m)				/* mean				*/
    {
	nstartff[j_par[0] - 1][0] = 1;
	startff[j_par[0] - 1][0][0] = d_par[0];
    } else				/* fixed effect			*/
	for (k = 0; k < nx_par; ++k)
	{
	    ++nstartff[j_par[0] - 1][j_par[1]];
	    startff[j_par[0] - 1][j_par[1]][k] = d_par[k];
	}
    return;
}
void set_start_var(int m)		/* sets the starting value(s)	*/
{					/* for variances		*/
    int j;
   
    for (j = 1; j <= nx_par; ++j)
	if (!m)				/* additive			*/
	    SIGSQ(j, j) = d_par[j - 1];
	else				/* residual			*/
	    TAUSQ(j, j) = d_par[j - 1];
    if (!m)
	naddvar = nx_par;
    else
	nresvar = nx_par;

    return;
}
void set_start_cor(int m)		/* sets the starting value(s)	*/
{					/* for correlations		*/
    int j, row, col;
   
    for (row = 2, j = 0; j < nx_par; ++row)
        for (col = 1; col < row && j < nx_par; ++col, ++j)
	    if (!m)			/* additive			*/
	        SIGSQ(row, col) = d_par[j];
	    else			/* residual			*/
	        TAUSQ(row, col) = d_par[j];
    if (!m)
    {
	naddcov = nx_par;
        add_cor_flag = 1;
    }
    else
    {
	nrescov = nx_par;
        res_cor_flag = 1;
    }
    return;
}
void set_start_cov(int m)		/* sets the starting value(s)	*/
{					/* for covariances		*/
    int j, row, col;
   
    for (row = 2, j = 0; j < nx_par; ++row)
        for (col = 1; col < row && j < nx_par; ++col, ++j)
	    if (!m)			/* additive			*/
	        SIGSQ(row, col) = d_par[j];
	    else			/* residual			*/
	        TAUSQ(row, col) = d_par[j];
    if (!m)
    {
	naddcov = nx_par;
        add_cor_flag = 0;
    }
    else
    {
	nrescov = nx_par;
        res_cor_flag = 0;
    }
    return;
}
void set_fit_cov(int m)			/* sets the fit/no-fit flag for	*/
{					/* covariances			*/
    int j, row, col; 

    for (row = 2, j = 0; j < nk_par; ++row)
        for (col = 1; col < row && j < nk_par;  ++col, ++j)
	    if (!m)			/* additive			*/
            {
                if (j_par[j] != 0)
		    SIG_MODL(row, col) = 1;
                else 
                    SIG_MODL(row, col) = 0;
            }
	    else			/* residual			*/	
 	    {
		if (j_par[j] != 0)		
		    TAU_MODL(row, col) = 1;
                else
		    TAU_MODL(row, col)  = 0;
            }
    if (!m)
	nfitadd = nk_par;
    else
	nfitres = nk_par;

    return;
}
void set_set_eigen(void)		/* allocates the array for the 	*/
{					/* eigenvalues and stores the 	*/
    int k;				/* starting values		*/

    peigen = malloc((nx_par + 1) * sizeof(double));
    alloc_chk(WHERE, peigen);
    for (k = 1; k <= nx_par; ++k)
	peigen[k] = d_par[k - 1];
    n_eigen = nx_par;

    return;
}
#include "proc_err.c"
#include "proc_drop.c"
#include "proc_auto.c"
#include "proc_seeds.c"
#include "proc_poly.c"
#include "proc_fils.c"
#include "check_pars.c"
#include "proc_end.c"
#include "input_eigen.c"

