#include "quant.h"

double lik_eigen(double **gmat, double *evec, int nn)
/* 
 * This function takes a general nn x nn dimension matrix gmat as input, 
 * and calls eispack function rg_() to solve the eigensystem.
 * It returns the largest eigenvalue and put the corresponding 
 * eigenvector in evec.               - March 15. 2001 N.L.
 */
{
    int i, j, ierr;
    int *iv1;
    double *dummy, *rvec, *ivec, *zz, *fv1;

    /* Request calculating eigenvectors */
    int calc_evec = 1;
    double max_evalue = -9999.9;
    int max_evalue_index = -1;

    dummy = malloc(nn * nn * sizeof(double));
    alloc_chk(WHERE, dummy);
    rvec = malloc(nn * sizeof(double));
    alloc_chk(WHERE, rvec);
    ivec = malloc(nn * sizeof(double));
    alloc_chk(WHERE, ivec); 
    zz = malloc(nn * nn * sizeof(double));
    alloc_chk(WHERE, zz);
    fv1 = malloc(nn * sizeof(double));
    alloc_chk(WHERE, fv1);
    iv1 = malloc(nn * sizeof(int));
    alloc_chk(WHERE, iv1); 
    
    /* Transpose the matrix into an array */
    for (j = 0; j < nn; ++j)
	for (i = 0; i < nn; ++i)
	    dummy[i + nn * j] = gmat[i][j];

    rg_(&nn, &nn, dummy, rvec, ivec, &calc_evec, zz, iv1, fv1, &ierr);

    /* The largest real eigenvalue and its index */
    for (i = 0; i < nn; ++i) {
	if ( rvec[i] > max_evalue && abs(ivec[i]) < 1.e-8 ) {
	    max_evalue = rvec[i];
	    max_evalue_index = i;
	}
    }

    /* The max_eigenvalue_index th row of zz is the corresponding
       eigenvector */
    for (i = 0; i < nn; ++i) 
	evec[i] = zz[i + max_evalue_index * nn];
    
    if (ChkEigC && ierr > 0)
        printf("\n error return code from rg_() is %d\n\n", ierr);
    
    free((void *) dummy);
    dummy = NULL;
    free((void *) rvec);
    rvec = NULL;
    free((void *) ivec);
    ivec = NULL;
    free((void *) zz);
    zz = NULL;
    free((void *) iv1);
    iv1 = NULL;
    free((void *) fv1);
    fv1 = NULL; 
    
    return max_evalue;
}
